#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Remote Model Calculation

last mod., Tue Oct 27 17:30:23 EDT 2020
'''
import sys
#sys.path.append("/apps/fafa/pyx/tst/")
#from __future__ import print_function
import pandas as pd, numpy as np, json, os, datetime
import simplejson
from pandas import Timestamp
from bson.objectid import ObjectId
from flask import render_template,Response,make_response
#from jinja2 import Template as jjTemplate
from types import *
from subprocess import Popen, PIPE
from glob import glob, iglob
from _alan_str import lsi2nlg_calc,combine_cmd2dict,jj_fmt,sysCall,find_mdb,write2mdb,dtCvt
from _alan_calc import subDict,sqlQuery,getKeyVal,pqint,subDF,subDict,renameDict
from _alan_date import tg_latest2week,tg_next2week,next_date,dt2ymd,ymd2dt,s2dt,utc2local
from alanapi import run_alanapi,opt_alanapi,search_comment,search_factor,search_quote,search_hist,search_history,search_list,search_allocation,search_mp3,search_ohlc
#,udfStr,roundUSD
from _alan_pppscf import batch_pppscf
from _alan_ohlc_fcs import batch_ohlc_fcs
from math import *
import requests
from importlib import import_module
if sys.version_info.major == 2:
	reload(sys)
	sys.setdefaultencoding('utf8')
pd.options.display.float_format='{:,.2f}'.format

pd.set_option('display.max_colwidth', -1)
pd.options.display.float_format = '{:,g}'.format

def pqrint(*args,**kwargs):
	return pqint(*args,**kwargs)

def rPost(url='',data={},timeout=5,ret='{}'):
	try:
		if len(url)<0:
			return ret
		r = requests.post(url, data=data,timeout=timeout)
		r.raise_for_status()
		if isinstance(r,dict):
			ret=r
		else:
			ret=r.content
	except requests.exceptions.HTTPError as e:
		sys.stderr.write("{}\n".format(e.response.text))
	return ret

def err_page(htmlName='errorhandler.html',code=-500,name="Invalid Calculation",description=''):
	''' return error message via html page
	'''
	return render_template(htmlName,code=code,name=name,description=description)

def display_page(pathname='index',request=None,**optx):
	return show_page(pathname=pathname,request=request,**optx)

def show_page(pathname='index',request=None,**optx):
	sys.stderr.write("===START show_page request:{}\n".format(request))
	opts=wrap_request(request)
	sys.stderr.write(" --opts:{}\n".format(opts))
	try:
		pageBase=pathname
		funcName='page_{}'.format(pageBase)
		htmlName=opts.pop('html',None)

		#----------------------------------------------------------------
		if funcName=='page_api' and htmlName is None:
			# Run 'page_api' command
			dd = page_api(**opts)
			if isinstance(dd,str):
				if opts.get('output','').lower() in ['html','csv'] and opts.get('outTF',False):
					return dd
				ret=Response(dd,mimetype='application/json')
				return ret
			ret=Response(simplejson.dumps(dd,ignore_nan=True),mimetype='application/json')
			return ret
		elif funcName in globals() and hasattr(globals()[funcName],'__call__'):
			# Run corresponding funcName like page_???()
			pass
		elif 'templates/{}'.format(htmlName) in glob('templates/*html'):
			# Newly added @ Sat Oct 24 21:25:43 EDT 2020
			# Process corresponding html page without funcName like page_???()
			ret = process_page(htmlName, **opts)
			return ret
		else:
			# Run home page page_index() as default
			sys.stderr.write("**WARNINGS:{} not found".format(funcName))
			funcName='page_{}'.format('index')
			sys.stderr.write(", use:{} instead".format(funcName))

		#----------------------------------------------------------------
		funcArg=globals()[funcName]

		sys.stderr.write(" --pathname:{}\n".format(pathname))
		sys.stderr.write(" --funcName:{}\n".format(funcName))
		sys.stderr.write(" --FuncObj:{}\n".format(funcArg))
		sys.stderr.write(" --show_page {} INPUT/opts:\n{}\n".format(funcName,opts))

		#----------------------------------------------------------------
		# Run corresponding command page_???
		dd = funcArg(**opts)

		sys.stderr.write(" --After run funcName:{},type:{}\n".format(funcName,type(dd)))
		sys.stderr.write(" --show_page OUTPUT:\n{}".format(dd)[:200]+"\n")

		#----------------------------------------------------------------

		# Use htmlName page is specified OR {pageBase}.html page
		if htmlName is None: 
			htmlName='{}.html'.format(pageBase)
		sys.stderr.write(" --htmlPage:{}\n".format(htmlName))

		# Use apitest.html as default page if htmlName page does not exist
		if 'templates/'+htmlName not in glob('templates/*html'):
			htmlName='apitest.html'

		# Process corresponding page
		ret = process_page(htmlName, dd, **opts)

		#----------------------------------------------------------------
	except Exception as e:
		sys.stderr.write("**ERROR: {} @ {}\n".format(str(e),"show_page"))
		#ret= str(e)
		ret= err_page(description=str(e)+"@ show_page")
	return ret

def process_page(htmlName, dd, **opts):
	# Processing output 
	from bson import json_util as json
	#opts.update({"import_module":import_module})
	#opts.update({"getattr":getattr})
	opts.update({"exists":os.path.exists})
	if isinstance(dd,dict):
		datax=json.dumps(dd,default=dtCvt)
		dd.update(datetime=datetime,pd=pd,s2dt=s2dt,utc2local=utc2local)
		ret= render_template(htmlName,datax=datax,**dd)
	elif isinstance(dd,list):
		opts.update(datetime=datetime,pd=pd,s2dt=s2dt,utc2local=utc2local)
		ret= render_template(htmlName,mtx=dd,**opts)
	else:
		opts.update(datetime=datetime,pd=pd,s2dt=s2dt,utc2local=utc2local)
		ret= render_template(htmlName, content=dd,**opts)

	sys.stderr.write("===processing page: {}\n".format(htmlName))
	sys.stderr.write(" --opts: {}".format(opts)[:60]+"\n")
	sys.stderr.write(" --data: {}".format(dd)[:200]+"\n")

	if isinstance(dd,dict) and 'submit' in dd: 
		ret = make_response(ret)
		if dd['submit'] == 'login':
			ret.set_cookie('userID', dd['user'])
		elif dd['submit'] == 'logout':
			ret.delete_cookie('userID')
		elif dd['submit'] == 'delete':
			ret.delete_cookie('userID')


	return ret

def page_api(**optx):
	opts={"search":"comment","outTF":False}
	opts.update(optx)
	dfTF = opts.pop('dfTF',False)
	opts.update(debugTF=True)
	pqint("===page_api INPUT:\n",opts, file=sys.stderr)
	data=wrap_alanapi(**opts)
	if isinstance(data,pd.DataFrame) and dfTF is False:
		dd=data.to_json(orient='records',force_ascii=False)
	elif isinstance(data,dict):
		dd={x:dtCvt(y) for x,y in data.items()}
	elif isinstance(data,list):
		for j in range(len(data)):
			if not isinstance(data[j],dict):
				continue
			for x,y in data[j].items():
				data[j][x]=dtCvt(y)
		dd=data
	else:
		dd=data
	return dd

def page_login(**optx):
	if 'submit' not in optx:
		return optx
	pqint(" -- page_login INPUT:\n",optx, file=sys.stderr)
	submit=optx.pop('submit','login')
	url = 'http://login.beyondbond.com/'
	if submit in ['logout','lost-password','reset-password','delete','signup']:
		url = url+submit
	dd = dict(submit=submit)
	try:
		ret = rPost(url=url,data=optx)
		userInfo= json.loads(ret.decode('utf-8'))
		pqint("==Authentication OUT:\n",userInfo, file=sys.stderr)
		if isinstance(userInfo,dict):
			dd.update(userInfo)
	except Exception as e:
		pqint("==WARNINGS Login:{}\n".format(str(e)), file=sys.stderr)
	pqint(" -- page_login OUTPUT:\n",dd, file=sys.stderr)
	return dd

def page_signup(**optx):
	return page_login(**optx)

def page_index(**optx):
	dd={'search': 'comment', 'output': 'dict','topic': 'hourly','outTF':False}
	optx.update(dd)
	#ret=wrap_alanapi(**optx)
	ret= page_api(**optx)
	if len(ret)>0:
		optx.update(ret[0])
	return optx

def page_report(**optx):
	topic,subtopic,search=getKeyVal(optx,['topic','subtopic','search'],['report',None,'comment'])
	optx.update(search=search,topic=topic,subtopic=subtopic,output='json')
	pqint(" --page_report INPUT:\n",optx, file=sys.stderr)
	if subtopic =='company':
		optx.update(topic='news')
	ret= page_api(**optx)
	pqint(" --page_report OUTPUT:\n",type(ret),"{}".format(ret)[:30],file=sys.stderr)
	if isinstance(ret,list) and len(ret)>0 and subtopic != 'company':
		return ret[0]
	else:
		return ret

def page_performance(**optx):
	optx.update(search='quote',topic='detailed',ticker='AAPL')
	ret = page_api(**optx)
	pqint(" --page_report OUTPUT:\n",type(ret),"{}".format(ret)[:30],file=sys.stderr)
	if isinstance(ret,list) and len(ret)>0: 
		return ret[0]
	else:
		return ret

def page_news(**optx):
	topic,subtopic,search=getKeyVal(optx,['topic','subtopic','search'],['news',None,'comment'])
	optx.update(search=search,topic=topic,subtopic=subtopic)
	return page_api(**optx)

def page_portfolio_OLD(**optx):
	opts=dict (search='comment',output='html',outTF=True)
	opts.update(optx)
	return page_api(**opts)

def page_portfolio(**optx):
	opts=dict (search='comment',dfTF=True)
	opts.update(optx)
	dd= page_api(**opts)
	if isinstance(dd,pd.DataFrame):
		cfm = {'marketCap': lambda x: "{:,.2f}B".format(x/10**9),'ytdRtn': "{:,.2%}".format, 
			'W.A.Return': "{:,.2%}".format,'AvgReturn': "{:,.2%}".format}
		clx=['values$MM','valuex$MM','PnL$MM']
		cfm.update({s : lambda x: "{:,.2f}".format(x/10**6) if not np.isnan(x) else "-" for s in clx})
		clx=['YTD%','forwardPE','trailingPE','close','xclose']
		cfm.update({s : lambda x: "{:,.2f}".format(x) if not np.isnan(x) else "-" for s in clx})
		cfm.update({'cDate': lambda x: "{}".format(int(x.strftime("%Y%m%d")))})
		ret = dd.to_html(index=False,formatters=cfm)
	else:
		ret = ''
	#allow adding a comment:cmt here, TBD @ Sat Oct 24 21:23:48 EDT 2020
	cmt=''
	ret = cmt + ret
	return ret

def page_globalmacro(**optx):
	return page_api(**optx)

def page_apitest(**optx):
	return page_api(**optx)

def wrap_request(request):
	dd={}
	pqint("===request METHOD: {}".format(request.method), file=sys.stderr)
	optGet = request.args.to_dict()
	pqint("===request GET Input:\t",optGet, file=sys.stderr)
	optPost = request.get_json()
	pqint("===request POST Input:\t",optPost,request.form, file=sys.stderr)
	user = request.cookies.get('userID')
	pqint("===request cookies user:{}".format(user) , file=sys.stderr)
	if request.method=='POST':
		if optGet is not None:
			dd.update(optGet)
		if optPost is not None:
			dd.update(optPost)
	else:
		if optPost is not None:
			dd.update(optPost)
		if optGet is not None:
			dd.update(optGet)
	if request.form is not None:
		dd.update(request.form.to_dict(flat=True))
	#if request.form is not None:
	#	dd.update(request.form)
	if user is not None:
		dd.update(user=user)
	pqint("===request REQUEST Input:\t{}".format(dd) , file=sys.stderr)
	return dd

def wrap_alanapi(**opts):
	try:
		dd, _ = opt_alanapi()
		dd.update(opts)
		fdLst=dd.pop('field',None)
		if 'ticker' in dd:
			tkLst=dd.pop('ticker','').split(',')
		else:
			tkLst=[]
		sys.stderr.write(" --alanapi INPUT:\n{}\n".format(dd))
		data=run_alanapi(tkLst,fdLst,**dd)
	except Exception as e:
		data=dict(err=str(e))
		sys.stderr.write("**ERROR:{} at wrap_alan\n".format(str(e)))
	return data

def run_page_alanapi(request):
	dd = wrap_request(request)
	return wrap_alanapi(**dd)

def find_lsi2nlg_list(findDct={'username':'system'},fieldLst=['tmplname'],dbname='ara',tablename='lsi2nlg'):
	if not isinstance(fieldLst,list):
		fieldLst = list(fieldLst)
	fieldDct = {x:1 for x in fieldLst}
	xg,_,_ = find_mdb(dbname=dbname,tablename=tablename,jobj=findDct,field=fieldDct)
	if len(xg)<1:
		return []
	else:
		df=pd.DataFrame(xg)
		ky = fieldLst[0]
		return df[ky].sort_values().unique()

def find_lsi2nlg_info(jobj={},fieldLst=['username','category','tmplname','lang'],dbname='ara',tablename='lsi2nlg'):
	findDct=subDict(jobj,fieldLst)
	mobj,clientM,err_msg = find_mdb(jobj=findDct,dbname=dbname,tablename=tablename)
	if len(mobj)>0:
		return (mobj[0])
	else:
		return []

def write_lsi2nlg_info(jobj,zpk=['username','category','tmplname','lang'],dbname='ara',tablename='lsi2nlg'):
	mobj,clientM,err_msg = write2mdb(jobj,dbname=dbname,tablename=tablename,zpk=zpk)
	return mobj

def argstr_tmplrpt(tmplstr, dux, *_xargs_):
	"""
	Update list of [_xargs_] command scripts to dict [dux] 
	and then apply [dux] to jinja2 template [tmplstr] and produce [tmplrpt]
	return text output [tmplrpt]
	"""
	try:
		duy=combine_cmd2dict(*_xargs_) # TBD:,_xdu_=dux)
		if duy:
			dux.update(duy)
		#dux.update({'udfStr':udfStr,'roundUSD':roundUSD})
		tmplrpt = jj_fmt(tmplstr,dux)
	except Exception as e:
		tmplrpt = str(e)
	return tmplrpt

def argstr_set(gd,excl_ky=[]):
	"""
	assign dict [gd] keys as variables exclude [excl_ky] key list
	"""
	argstr=''
	for j,(ky,va) in enumerate(gd.iteritems()):
		if ky in excl_ky:
			continue
		if isinstance(va,StringTypes) and ky != "ts":
			argstr += "{}='{}';\n".format(ky,va)
		elif isinstance(va, (float, int, list, dict, tuple)):
			argstr += "{}={};\n".format(ky,va)
	return argstr

def run_j2ts(optGet={},optPost={}):
	pqint(optGet,optPost, file=sys.stderr)
	dd = {}
	if optPost is not None:
		dd.update(optPost)
	if optGet is not None:
		dd.update(optGet)
	if 'j2ts' in dd:
		j2ts = dd['j2ts'] 
		dd = subDict(dd,['j2ts'],reverseTF=True)
	else:
		j2ts = 'Usage of /?key1=value1&key2=value2 ...'
	return jj_fmt(j2ts,dd,j2onlyTF=True)

def tk_video(ticker, lang):
	xcmd = ('./stock_ticker2video.sh {} {}').format(ticker, lang)
	pqint('RUNNING ', xcmd, file=sys.stderr)
	p = Popen(xcmd, shell=True, bufsize=1024, stderr=PIPE, stdout=PIPE)
	try:
		outs, errs = p.communicate()
		return True
	except:
		p.kill()
		outs, errs = p.communicate(timeout=30)

	return False

def upd_video(name, lang):
	vdLst = sorted(iglob(('static/{}_{}.*.mp4').format(name, lang)), key=os.path.getmtime, reverse=True)
	# create the video if it does not exists 
	if len(vdLst) < 1:
		upd_flg = True
	else:
		fname = vdLst[0]
		xtime = datetime.datetime.fromtimestamp(float(fname.split('.')[-2:][0]))
		ctime = datetime.datetime.now()
		# re-create video for not today
		if ctime.strftime('%Y%m%d') != xtime.strftime('%Y%m%d'):
			upd_flg = True
		else: 
			# re-create video for older than 6-hour
			if (ctime - xtime).seconds / 3600 >= 12:
				upd_flg = True
			else:
				upd_flg = False
	if upd_flg is True:
		TF = tk_video(name, lang)
		vdLst = sorted(iglob(("static/{}_{}.*.mp4").format(name, lang)), key=os.path.getmtime, reverse=True)
		if len(vdLst) < 1:
			pqint("**ERROR: {} @ {}".format(fname,'stock_ticker2video','upd_video()'), file=sys.stderr)
			return ''
		fname = vdLst[0]
		for xf in vdLst[2:]:
			os.remove(xf)

	else:
		pqint(("\t{} exist @ {}").format(fname, xtime), file=sys.stderr)
	return os.path.basename(fname)


def qt_result(name,lang,videoYN='0',errMsg=''):
	from yh_chart import yh_hist_query as yhq
	#from dateutil.parser import parse as Timestamp
	name=str(name.upper())
	videoYN=str(videoYN)
	d = dict(name=name,lang=lang,videoYN=videoYN,len=len,enumerate=enumerate,error_message=errMsg)
	if videoYN == '1':
		pqint("\tvideoYN:{} @ {}".format(videoYN, 'qt_result'), file=sys.stderr)
		try:
			e=''
			fname = upd_video(name, lang)
		except Exception as e:
			fname=''
		if len(fname)<1:
			errMsg = "**ERROR: No [{}_{}] video {} @ {}".format(name,lang,str(e),'upd_video')
		d.update(error_message=errMsg,mp4path=fname)
		return d
	pqint("\tvideoYN:{} @ {}".format(videoYN, 'qt_result'), file=sys.stderr)
	try:
		qt = yhq([name],types='quote',rawTF=True,dbname=None)[0]
		pqint("===QT:\n{}".format(qt), file=sys.stderr)
	except Exception as e:
		pqint("===yh_hist_query: {}".format(str(e)), file=sys.stderr)
		qt = {}
	if 'regularMarketTime' in qt:
		qt['lastUpdate'] = datetime.datetime.fromtimestamp(qt['regularMarketTime']).strftime("%Y/%m/%d %H:%M:%S")
	d.update(qt)
	try:
		d['quote'] = render_template('tp_quote.html',quote=qt)
	except Exception as e:
		pqint("===render_template: {}".format(str(e)), file=sys.stderr)
	if 'longName' in qt:
		d['company'] = qt['longName']
	if 'epsTrailingTwelveMonths' in qt:
		d['earnings'] = qt['epsTrailingTwelveMonths']
	if 'trailingPE' in qt:
		d['financials'] = qt['trailingPE']
	return d

def run_rmc(jobj):
	"""
	run Remote-Market-Commentary based on input dict: [jobj]
	with keys: tmplstr,argstr,ticker,tmplrpt,username,category,tmplname,action
	return string for Market Commentary
	"""
	global username, ticker, lang , dirname
	if 'username' in jobj:
		username=jobj['username']
	if 'ticker' in jobj:
		ticker=jobj['ticker']
	if 'lang' in jobj:
		lang=jobj['lang']
	dirname='templates/'
	jobj.update(dirname=dirname)
	jobj.update(mp3YN=False)
	jobj.update(start=dt2ymd(next_date()))
	# run rmc based on tmplstr
	if str(jobj['tmpltype']) == 'pppscf' and str(jobj['action']) == 'run':
		jobj['tmplrpt'] = batch_pppscf([jobj['ticker']],optx={"j2ts":jobj['tmplstr'],"lang":jobj['lang'],"debug":True,"saveDB":False,"category":jobj['category']})
		# run_alan_pppscf.batch_pppscf() 1st then import gData in batch_pppscf()
		from _alan_pppscf import gData
		jobj["retcode"]="run successfully"
	elif str(jobj['tmpltype']) == 'fcs' and str(jobj['action']) == 'run':
		jobj['tmplrpt'] = batch_ohlc_fcs([jobj['ticker']],optx={"j2ts":jobj['tmplstr'],"lang":jobj['lang'],"debug":True,"saveDB":False,"category":jobj['category']})
		from _alan_ohlc_fcs import gData
		jobj["retcode"]="run successfully"
	elif str(jobj['tmpltype']) == 'eps' and str(jobj['action']) == 'run':
		from fcs_eps_price import run_eps_fcs
		gData = run_eps_fcs([jobj['ticker']])[0]
		jobj['tmplrpt'] = argstr_tmplrpt(jobj['tmplstr'],gData,jobj['argstr'].replace("\n",''))
		jobj["retcode"]="run successfully"
	elif str(jobj['tmpltype']) == 'mkt' and str(jobj['action']) == 'run':
		# Same as command line:
		# headline_writer.py --lang=cn --extra_xs='onTheFly=True;j2ts="{% include \"daily_briefing_cn.j2\" %}";dirname="templates"' --start=20190705 2>/dev/null

		from headline_writer import generate_headline
		optx=dict(j2ts=jobj['tmplstr'],onTheFly=True)
		optx.update(jobj)
		if jobj['argsflg']=='1' and len(jobj['argstr'])>0:
			gdd={}
			exec(jobj['argstr'],globals(),gdd)
			optx.update(gdd)
		try:
			ret = generate_headline(**optx)
		except Exception as e:
			err_msg = "**ERROR:{} @ {} ".format(str(e),"generate_headline")
			pqint(err_msg, file=sys.stderr)
			ret=err_msg
		jobj['tmplrpt'] = ret
	elif str(jobj['tmpltype']) == 'test' and str(jobj['action']) == 'run':
		optx={'j2ts_header':jobj['tmplstr']}
		optx.update(jobj) 
		try:
			# use 'jobj' as internal object and update jobj's key/val into **optx 
			pqint("===username: {}, argflg: {},argstr: {}".format(username,jobj['argsflg'],jobj['argstr']),file=sys.stderr)
			if jobj['argsflg']=='1' and len(jobj['argstr'])>0:
				gdd={}
				xstr = jobj['argstr'] 
				exec(xstr,globals(),gdd)
				if 'jobj' in gdd:
					optx.update(gdd['jobj'])
					pqint("Add additional:{}".format(gdd['jobj']),file=sys.stderr)
				else:
					optx.update(gdd)
					pqint(gdd,file=sys.stderr)
		except Exception as e:
			pqint("**ERROR:{} @ {} ".format(str(e),"exec argstr"), file=sys.stderr)
		pqint(optx)
		ret = jj_fmt(jobj['tmplstr'],**optx)
		jobj['tmplrpt'] = ret
	# TBD: testing 'get' instead of lsi2nlg_calc()
	elif str(jobj['action']) == 'get':
		mobj = find_lsi2nlg_info(jobj=jobj)
		jobj.update(mobj)
		jobj['argsflg']=1
	# TBD: testing 'save' instead of lsi2nlg_calc()
	#elif str(jobj['action']) == 'save':
	#	mobj = write_lsi2nlg_info(jobj)
	#	jobj['argsflg']=1
	else:
		jobj['argstr'] = jobj['argstr'].replace("\n",'')
		jobj = lsi2nlg_calc(jobj)
		jobj['argsflg']=1
	# run rmc with additional argstr command
	if str(jobj['tmpltype']) in ['eps','fcs','pppscf'] and str(jobj['action']) == 'run':
		if 'itnstr' in jobj:
			try:
				jobj['itnstr'] = argstr_set(gData,['fp'])
			except Exception as e:
				pqint("**ERROR:{} @ {} ".format(str(e),"Assign Internal Variables"), file=sys.stderr)
		try:
			if 'argsflg' in jobj and int(jobj['argsflg'])==1:	
				jobj['tmplrpt'] = argstr_tmplrpt(jobj['tmplstr'],gData,jobj['itnstr'],jobj['argstr'].replace("\n",''))
		except Exception as e:
			pqint("**ERROR:{} @ {} ".format(str(e),"Apply Additional Variables"), file=sys.stderr)
			jobj["retcode"]=str(e)
	jobj["lastmod"]=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	jstr=json.dumps(jobj)
	return jstr

def topic_theme_etffund(**opts):
	from etf_holdings_ytd import etf_holdings_ytd as ehy
	fund=getKeyVal(opts,'fund','XLK')
	df = ehy(etfname=fund)
	colz=['ticker','weight%','YTD%','close','values$MM','valuex','pnl','iDate','fund','date','company']
	colx = [x for x in colz if x in df.columns]
	df=df[colx]
	df=renameDict(df,{"valuex":"valuex$MM","pnl":"PnL$MM","date":"cDate"})
	sys.stderr.write("==DF of {} from topic_theme_etffund():\n{}\n".format(fund,df))
	return df

def topic_theme_majorplayer(**opts):
	'''
	use 'themename' as tablename
	'''
	from _alan_str import find_mdb
	fund=getKeyVal(opts,'fund','renaissance-technologies-llc')
	clientM=getKeyVal(opts,'clientM',None)
	dbname=getKeyVal(opts,'dbname','ara')
	themename=getKeyVal(opts,'themename','topic_theme_majorplayer')
	dfTF=getKeyVal(opts,'dfTF',True)
	jobj = dict(fund=fund)
	data0,mDB,errmsg = find_mdb(jobj,tablename=themename,dbname=dbname,sortLst={"funddate"},limit=1,dfTF=False)
	mxdate=data0[0]['funddate']
	jobj.update({"funddate":mxdate})
	df,mDB,errmsg=find_mdb(jobj,tablename=themename,dbname=dbname,dfTF=dfTF)
	colX = ['ticker','close','pbdate','marketCap',
		'changePercent', 'CurrentShares','percentPos','SharesChangePercent',
		'fiftyTwoWeekRange','shortName',
		'fund','funddate','pbdt']
	colDc = {'close':'closePrice','pbdate':'closeDate','shortName':'company',
		'SharesChangePercent':'SharesChg%','changePercent':'Chg%',
		'percentPos':'Position%'
		}
	df=subDict(df,colX)
	df=renameDict(df,colDc)
	df = df.dropna(subset=['closeDate'])
	df['closeDate'] = [int(x) if not np.isnan(x) else 0 for x in df['closeDate'].values]
	return df

def topic_theme_media(**opts):
	from _alan_str import find_mdb
	jobj = {"buyCount":{"$gt":5}}
	dtmp,mDB,errmsg=find_mdb(jobj,tablename='topic_theme_media',dbname='ara',dfTF=True)
	df = dtmp.sort_values(by='buyCount',ascending=False)
	colX = ['ticker','buyDate','marketCap','buyPrice','closePrice','buyChg%','buyCount','dayChg%','EPS','pbdate','Range52Week','Company','sector','industry','pubDate']
	if 'marketCap' in df:
		df['marketCap'] = df['marketCap'].astype(float)
	df=subDF(df,colX)
	return df

def topic_theme_media_OLD(**opts):
	from _alan_str import find_mdb
	subtab=getKeyVal(opts,'subtab','blacklist')
	if subtab.lower() == 'whitelist':
		jobj = {"rrt":{"$gt":10}}
		dtmp,mDB,errmsg=find_mdb(jobj,tablename='topic_theme_media',dbname='ara',dfTF=True)
		df = dtmp.sort_values(by='rrt',ascending=False).iloc[:100]
	elif subtab.lower() == 'conflict':
		dtmp,mDB,errmsg=find_mdb({},tablename='topic_theme_media',dbname='ara',dfTF=True)
		a1=list(dtmp.query("rrt>=1.0")['ticker'].unique())
		a2=list(dtmp.query("rrt<=-1.0")['ticker'].unique())
		aLst=set(a1).intersection(set(a2))
		df = dtmp.loc[dtmp['ticker'].isin(aLst)].sort_values(by='ticker')
	else:
		jobj = {"rrt":{"$lt":-10}}
		dtmp,mDB,errmsg=find_mdb(jobj,tablename='topic_theme_media',dbname='ara',dfTF=True)
		df = dtmp.sort_values(by='pbdate',ascending=False).iloc[:100]
		colX=["ticker","pbdate","start","end","sPrice","ePrice","rrt","Price","SegmentDscript","CallDscript","Company"]
		df=subDF(df,colX)
	return df

def topic_theme_ipo(**opts):
	from _alan_str import find_mdb
	subtab=getKeyVal(opts,'subtab','')
	opts.pop('output',None) # not output type assigned
	updTF=opts.pop('updTF',False)
	jobj=json.loads(opts.pop('jobj','{}'))
	sector=opts.pop('sector','')
	if len(sector)>5:
		sLst=sector.split(',')
		jobj.update(sector={"$in":sLst})

	#df = get_ipoReport(updTF=updTF,**opts) # DEPRECATED
	#df = create_topic_theme_ipo(updTF=updTF,**opts) # SETUP in crontab

	tablename=opts.pop('tablename',None)
	if tablename is None:
		tablename='topic_theme_ipo'
	df,_,_ = find_mdb(jobj,dbname='ara',tablename=tablename,dfTF=True)
	colX = ['ticker','ipoDate','marketCap','ipoPrice','Price','closePrice','ipoChg%','dayChg%','EPS','Range52Week','Company','sector','industry','pubDate']
	df=subDict(df,colX)

	#-onTheFly run, not used
	#-run daily crontab to create 'topic_theme_ipo' table via  
	#-python3 -c "from ipoReport import create_topic_theme_ipo as ctt;ctt()"

	return df

def run_topic_theme(dfTF=True,**opts):
	subtopic=getKeyVal(opts,'subtopic','majorplayer')
	funcName = "_".join(["topic","theme",subtopic]) 
	funcArg = globals()[funcName] if funcName in globals() else topic_theme_majorplayer
	try:
		pqint(" --dfTF:{},{},funcArg:{} @{}".format(dfTF,funcArg,opts,'run_topic_theme()'),file=sys.stderr)
		df = funcArg(**opts)
		if dfTF:
			return df
		cfm = {'marketCap': "{:,.0f}".format}
		ret = df.to_html(index=False,formatters=cfm)
		sty = "<style>.dataframe {text-align:right;}</style>\n"
		return sty+ret
	except Exception as e:
		pqint(str(e))
		return str(e)

def run_api(jobj,engine=None):
	pd.options.display.float_format = '{:,.2f}'.format
	sty = "<style>.dataframe {text-align:right;}</style>\n"
	ret="""Usage of:
	<H3>
	?topic='TOPIC'
	</H3>
	Where
	<PRE>
	TOPIC = [ipo|theme|priceRange|top10|utdEarningsList|mongo_search|file|write2file|test] """
	topic=getKeyVal(jobj,'topic','').lower()
	if topic is None:
		return ret

	if topic == 'theme':
		subtopic=getKeyVal(jobj,'subtopic','majorplayer')

	if topic == 'theme' and subtopic in ['majorplayer','media','ipo','etffund']:
		jobj.update(subtopic=subtopic)
		return run_topic_theme(dfTF=False,**jobj)
	elif topic == 'pricerange':
		xqTmp= '''select * from (select ticker,close as price,"trailingPE" as "peRatio","marketCap"::float/1000000 as "marketCapMM","changePercent" as "change%%",change from yh_quote_curr where close>={} and close<={}) as a, (select ticker,company_cn,company,sector_cn from mapping_ticker_cik where act_code>0) as b where a.ticker=b.ticker order by price'''
		try:
			vr=jobj['range'].split(",") if 'range' in jobj else [60,70]
			if len(vr)>=2:
				vr=np.array(vr,dtype=float)
			lb,ub = (vr.min(),vr.max())
			xqr = xqTmp.format(lb,ub)
			df = sqlQuery(xqr)
			pd.options.display.float_format = '{:,.2f}'.format
			cfm = {'marketCapMM': "{:,.0f}".format}
			ret = df.to_html(formatters=cfm)
			return sty+ret
		except Exception as e:
			pqint(str(e))
			return str(e)
	elif topic == 'top10':
		return run_top10(jobj,engine=None)
	elif topic in ['utd_earnings_list','utdearningslist']:
		from utd_earnings_list import utd_earnings_list
		dd=dict(sector='Technology',pbdate='20190101')
		dd.update(subDict(jobj,['sector','pbdate']))
		df=utd_earnings_list(**dd).sort_values(by=['pbdate','marketCap'],ascending=[False,False])
		df.rename(columns={'quarter':'epochs'},inplace=True)
		cfm = {'marketCap': "{:,.0f}".format,'recdate':"{:.0f}".format,'epochs':"{:.0f}".format }
		ret = df.to_html(formatters=cfm)
		return sty+ret
	elif topic in ['daily_med','dailymed']:
		from dailyMed_api import drug2labelInfo as dli
		try:
			dd =  dli(jobj['drug_name'])
		except Exception as e:
			sys.stderr.write("**ERROR:{}\n".format(str(e)))
			return(str(e))
		#return dd
		ts="""{{drug_name}}
		<P>
		{{ sec_cn|join('</P><P>\n') }}
		</P>
		"""
		return jj_fmt(ts,**dd)
	elif topic == 'mongo_search':
		d=dict(dbname='ara',tablename='lsi2nlg',username='system',field='tmplname',ticker='AAPL')
		if len(jobj)>0:
			d.update(jobj)
		findDct={'username':d.get('username')}
		fieldLst = d.get('field').split(',')
		fieldDct = {x:1 for x in fieldLst}
		xg,_,_ = find_mdb(dbname=d['dbname'],tablename=d['tablename'],jobj=findDct,field=fieldDct)
		return xg
	elif topic == 'file':
		try:
			dirname = jobj['dirname'] if 'dirname' in jobj else "/apps/fafa/pyx/tst"
			if 'j2name' in jobj:
				fname= "{}/{}".format(dirname,jobj['j2name'])
				ret = open(fname).read()
			elif 'image' in jobj:
				fname= "{}/{}".format(dirname,jobj['image'])
				ret = open(fname).read()
		except Exception as e:
			sys.stderr.write("**ERROR:{}".format(str(e)))
	elif topic == 'write2file':
		# save 'j2name' to 'dirName'  
		pqint(jobj)
		try:
			dirName = "/apps/fafa/pyx/flask/rmc/templates"
			if 'j2name' in jobj and 'j2ts' in jobj:
				if 'dirname' in jobj:
					dirName = jobj['dirname']
				fname = "{}/{}".format(dirName,jobj['j2name'])
				fp = open(fname,'w+')
				fp.write(jobj['j2ts'])
				fp.close()
				pqint("===Save {} To {}".format(jobj['j2ts'],fname))
				ret = "Successfully save to {}".format(fname)
		except Exception as e:
			sys.stderr.write("**ERROR:{}".format(str(e)))
			ret = str(e)
	elif topic == 'test':
		if 'tmplrpt' in jobj:
			ret= jobj['tmplrpt']
	return ret

def run_top10(jobj,engine=None):
	xobj={}
	if jobj['category'] == 'backtest':
		xql = "SELECT * FROM ara_ranking_backtest_cn"
		df = sqlQuery(xql,engine=engine)
		xobj = df.to_dict(orient='records')
	return xobj



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Usage of:
python3 -c "from plot_templates import plot_ts as pts;pts('^GSPC', title='標普500',chartname='',backend='tkAgg',debugTF=True)"
OR (for daily market index historical performance)
python3 -c "from plot_templates import plot_intraday_headline as pts;optx=dict(ranged='20210101,',gap='1d',xaxis='pbdate',rsiYN=True,pchgTF=2,x_fmt='%m/%d/%y');pts(chartname='',backend='tkAgg',debugTF=True,**optx)"
"""

import math
import sys
import pandas as pd
import numpy as np
import re
import matplotlib
matplotlib.use('agg')
if matplotlib.get_backend().lower() != 'tkagg':
	matplotlib.use('Agg', force=True)
import seaborn as sns
from matplotlib.ticker import FuncFormatter
from matplotlib.ticker import StrMethodFormatter
import matplotlib.dates as mdates
from datetime import datetime, timedelta, time
from _alan_calc import getKeyVal
from _alan_date import next_date,dt2ymd
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()
import matplotlib.pyplot as plt
import matplotlib.font_manager as mfm
font_path = "/usr/share/fonts/truetype/arphic/uming.ttc"
def fontProp(fname=font_path,**optx): 
	return mfm.FontProperties(fname=fname,**optx)
prop = fontProp()
colorLst=['blue','salmon','greenyellow','red','lightgray','cyan']

#plt.style.use('classic')
if sys.version_info.major == 2:
	reload(sys)
	sys.setdefaultencoding('utf8')

def gen_chartpath(chartname='',chartdir='US/mp3_hourly',chartformat='svg',epoch=None,**optx):
	"""
	Generate 'chartpath', 'chartname' as
	  {chartdir}/{templateName}_{ticker}_{epoch}.{chartformat},
	  {templateName}_{ticker}_{epoch}.{chartformat}
	where
	  'chartdir','chartformat' and 'epoch' are optional
	Example,
	  gen_chartpath(templateName='stock_performance',ticker='AAPL')
	Note,
	  'ticker' with '^' or '=' characters will be removed
	"""
	chartname='' if not isinstance(chartname,str) else chartname
	if len(chartname)>0:
		if len(chartdir)>0:
			chartpath="{}/{}".format(chartdir,chartname)
		else:
			chartpath = chartname
		return chartpath,chartname
	if epoch is None:
		epoch = dt2ymd(next_date(),"%s")
	chartname = chartpath = ''
	ky=['templateName','ticker']
	kLst=getKeyVal(optx,ky,[],extendTF=True)
	kLst.extend([epoch,chartformat])
	if not any([ x is None for x in kLst]):
		chartname = "{}_{}_{}.{}".format(*kLst)
		chartname = re.sub('[\^=]','',chartname)
	if len(chartdir)>0 and len(chartname)>0:
		chartpath = "{}/{}".format(chartdir,chartname)
	else:
		chartpath = chartname
	return chartpath,chartname

def get_data(ticker,dv=[],src='yh',**optx):
	from _alan_calc import pullStockHistory as psd
	ranged = getKeyVal(optx,'ranged','1d')
	gap = getKeyVal(optx,'gap','1m')
	datax = psd(ticker,gap=gap,ranged=ranged,days=1500,src=src)
	df = datax
	print("===>",type(df),df.tail())
	if 'volume' in datax:
		dv = datax[['volume']]
	return (df,dv)

def plot_daily_conclusion(da,db=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	'''
	plot daily_peers by marketcap barh
	'''
	sys.stderr.write("===start:{}()\n".format("plot_daily_conclusion"))
	ticker = getKeyVal(optx,'ticker')
	if backend is not None:
		plt.switch_backend(backend)
	chartpath,chartname = gen_chartpath(**optx)
	sns.set(style="whitegrid")

	# define plot
	if fig is None:
		fig, ax = plt.subplots(figsize=figsize)
	ax.grid(False, axis='y')
	ax.spines['right'].set_color('none')
	ax.spines['top'].set_color('none')
	ax.spines['bottom'].set_color('none')
	ax.spines['left'].set_color('none')

	bottom=optx['bottom'] if 'bottom' in optx else 0.25

	#da.plot(ax=ax,kind=kind,position=position,width=width)
	da = da[['comment','pchg']].sort_values(by='pchg',ascending=False).iloc[0:10,:]
	plt.title('指標變化百分比',fontproperties=fontProp(size=15))
	da['comment'] = [ xstr.strip('[,.，。]') for xstr in da['comment'].values ]
	ax.barh(y=da['comment'], width=da['pchg'])
	ax.set_yticklabels(labels=da['comment'],fontproperties=fontProp(size=15,weight='bold'))
	# plot features
	loc=optx['loc'] if 'loc' in optx else 9
	ncol=optx['ncol'] if 'ncol' in optx else 2
	fancybox=optx['fancybox'] if 'fancybox' in optx else True
	borderaxespad=optx['borderaxespad'] if 'borderaxespad' in optx else 0.
	plt.legend(loc=loc, ncol=ncol, fancybox=fancybox, borderaxespad=borderaxespad)
	rotation=optx['rotation'] if 'rotation' in optx else '30'
	fontsize=optx['fontsize'] if 'fontsize' in optx else 12
	plt.xticks(rotation=rotation,fontsize=fontsize)
	plt.legend(frameon=False)
	if len(title)>0:
		plt.title(title)
	plt.subplots_adjust(left=0.65,bottom=bottom)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return chartpath,ax,fig

def plot_mktidx(tkLst=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	sys.stderr.write("===start:{}()\n".format("plot_mktidx"))
	chartpath = getKeyVal(optx,'chartpath','')
	optx=locals()
	from csv2plotj2ts import get_csv2plot
	optx.update(optx['optx'])
	optx.pop('optx',None)
	opts ={'debugTF': True,'pngname': None, 'debugTF': True, 'rsiYN': True, 'renColumns': {'^IXIC': '納斯達克指數', '^DJI': '道瓊指數', '^GSPC': '標普500'}, 'colLst': 'close,pchg,epochs,ticker', 'hdrLst': None, 'npar': 15, 'days': 730, 'j2ts': None, 'start': None, 'pivot_group': 'ticker', 'gap': '1m', 'trendTF': False, 'ohlcTF': False, 'src': 'yh', 'pivot_value': 'close', 'end': None, 'title': '美股大盤走勢2019-09-17', 'lang': 'cn', 'backend': 'tkAgg', 'nbins': 6, 'tsTF': True, 'ranged': '1d', 'ohlcComboTF': False, 'sep': '|', 'interpolateYN': False, 'xaxis': 'epochs', 'pltStyle': 'classic', 'x_fmt': '%H:%M'}
	if len(tkLst)<1:
		tkLst= ['^GSPC', '^DJI', '^IXIC']
	plt.switch_backend(backend)
	if len(chartpath)>5:
		optx.update(pngname=chartpath)
	datax,fig,ax = get_csv2plot(tkLst,opts=opts,**optx)
	return(chartpath,ax,fig)

def plot_intraday_headline(tkLst=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	if backend is not None and len(backend)>1:
		plt.switch_backend(backend)
	sys.stderr.write("===start:{}()\n".format("plot_intraday_headline"))
	bottom=optx['bottom'] if 'bottom' in optx else 0.2
	chartpath = getKeyVal(optx,'chartpath','')
	chartname = getKeyVal(optx,'chartname','')    
	x_fmt = getKeyVal(optx,'x_fmt','%H:%M')    
	#chartpath,chartname = gen_chartpath(**optx)    
	from csv2plotj2ts import get_csv2plot,get_csvdata,subDict
	opts ={'debugTF': True,'pngname': None, 'debugTF': True, 'rsiYN': True, 'renColumns': {'^IXIC': '納斯達克指數', '^DJI': '道瓊指數', '^GSPC': '標普500'}, 'colLst': 'change,pchg,close,epochs,ticker', 'hdrLst': None, 'npar': 15, 'days': 730, 'j2ts': None, 'start': None, 'pivot_group': 'ticker', 'gap': '1m', 'trendTF': False, 'ohlcTF': False, 'src': 'yh', 'pivot_value': 'close', 'pchgTF':True, 'end': None, 'title': '美股大盤走勢2019-09-17', 'lang': 'cn', 'backend': 'tkAgg', 'nbins': 6, 'tsTF': True, 'ranged': '1d', 'ohlcComboTF': False, 'sep': '|', 'interpolateYN': False, 'xaxis': 'epochs', 'pltStyle': 'classic', 'x_fmt': '%H:%M'}
	if len(tkLst)<1:
		tkLst= ['^GSPC', '^DJI', '^IXIC']
	if len(chartpath)>5:
		optx.update(pngname=chartpath)
	opts.update(**optx)
	renColumns = opts['renColumns']
	gkys=['sep','src','days','start','end','colLst','hdrLst','debugTF','ranged','gap','tsTF','pchgTF','searchDB']
	#df = get_csvdata(tkLst, **subDict(opts,gkys))
	df = get_csvdata(tkLst, **opts)
	if df is None or len(df)<1 or df.size<1:
		return '',{},{}
	sys.stderr.write(" --opts:{},DF:\n{}".format(opts,df.tail().to_string()) )    
	#exit(1)
	
	da=pd.DataFrame()
	dc=pd.DataFrame()
	for tk in tkLst:
		da[tk] = df[df['ticker']==tk]['pchg']    
		dc[tk] = df[df['ticker']==tk]['change']    
	#da = (da.pct_change().fillna(0)+1).cumprod()   
    
	if debugTF:
		sys.stderr.write("\n===== da:\n{}\n\t@{}\n".format(da.to_csv(),'plot_intraday_headline'))
	
	max_num = da.values.max()
	min_num = da.values.min()
	min_index = np.min(da.index)

	try:
		plt.rcParams.update(plt.rcParamsDefault)
		sns.set_style("whitegrid",{'grid.linestyle':'--'})
		fig, ax=plt.subplots(figsize=figsize)
		fig.autofmt_xdate()
		if(opts['gap'][-1:]=='m'):
			x_lim = min_index.replace(hour=16, minute =0 )    
			plt.xlim(min_index-timedelta(minutes=5), x_lim+timedelta(minutes=5))
		##plt.ylim(min_num-(max_num-min_num)/2, max_num+(max_num-min_num)/5)
		#plt.xlabel('Time',fontsize=15)
		#plt.ylabel('price',fontsize=15)
		#hours = mdates.HourLocator(interval = 1)
		h_fmt = mdates.DateFormatter(x_fmt)
		#ax.xaxis.set_major_locator(hours)
		ax.xaxis.set_major_formatter(h_fmt)
		ax.grid(False, axis='x')                        
		ax.spines['right'].set_color('none')
		ax.spines['top'].set_color('none')
		ax.spines['bottom'].set_color('none')
		ax.spines['left'].set_color('none')
		#ax.set_ylabel('Returns Since Previous Close in %')
		ax.set_ylabel('自前日收盤價報酬率%',fontproperties=fontProp())
		plt.title(title,fontproperties=fontProp(size=15))
		#laLst=('標普500','道瓊指數','納斯達克指數')
		#for j,xlabel in enumerate(laLst):
		for j,tkX in enumerate(tkLst):
			xlabel = renColumns[tkX]
			xlev=df.loc[df['ticker']==tkX,'close'].dropna().iloc[-1] 
			xchg=da[tkLst[j]].dropna().iloc[-1]
			xchange=dc[tkLst[j]].dropna().iloc[-1]
			xa = "{:s}: {:,.0f} ({:+,.0f}, {:,.2%})".format(xlabel,xlev,xchange,xchg)
			plt.plot(da.index, da[tkX],label=xa,color=colorLst[j])
        
		ax.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,.2%}'.format(x)))
		plt.legend(prop=fontProp())        
		if debugTF:
			sys.stderr.write("===plot data:\n {}\n".format(df.tail(2)))
			sys.stderr.write("===plot pchg:\n {}\n".format(da.tail(2)))
		plt.subplots_adjust(left=0.1,bottom=bottom)
		if chartname is not None and len(chartname)>5:
			plt.savefig(chartpath, format=chartformat) #, bbox_inches='tight',dpi=1000)
		elif backend is not None and backend.lower()=='tkagg': 
			plt.show()
	except Exception as e:
		sys.stderr.write("**ERROR: {} @{}\n".format(str(e),"plot_intraday_headline"))
		return '',None,None   
	return(chartpath,ax,fig)


def plot_daily_macro(df,dv=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=True,**optx):
	if len(df)<1:
		return '',None,None
	sys.stderr.write("===start:{}()\n".format("plot_daily_macro"))
	src = getKeyVal(optx,'src','fred')
	if not isinstance(df,pd.DataFrame):
		ticker=df
		from _alan_calc import pullStockHistory as psd
		ranged = getKeyVal(optx,'ranged','1d')
		gap = getKeyVal(optx,'gap','1m')
		datax = psd(ticker,gap=gap,ranged=ranged,days=1500,src=src)
		datax.rename(columns={'close':ticker}, inplace=True)
		df = datax[[ticker]]
		if src=='fred':
			tk2=ticker+'_PCTCHG'
			dx2 = psd(tk2,src=src,days=1500)
			if len(dx2)>0:
				dv = datax[[ticker]]
				df = dx2[['close']]
		elif 'volume' in datax:
			dv = datax[['volume']]
	if 'volume' in df:
		dv = df[['volume']].copy()
		cs = list(df.columns)
		cs.pop(cs.index('volume'))
		df = df[cs]
	if backend is not None and len(backend)>1:
		plt.switch_backend(backend)
	trendTF = getKeyVal(optx,'trendTF',True)
	bottom=optx['bottom'] if 'bottom' in optx else 0.35
	chartpath,chartname = gen_chartpath(**optx)
	try:
		from _alan_pppscf import vertex_locator
		if not isinstance(df,pd.DataFrame):
			df = pd.DataFrame(data=df)
		if fig is None:
			plt.rcParams.update(plt.rcParamsDefault)            
			sns.set_style("whitegrid",{'grid.linestyle':'--'})                 
			fig, ax=plt.subplots(figsize=figsize)
			ax.grid(False, axis='x')                        
			ax.spines['right'].set_color('none')
			ax.spines['top'].set_color('none')
			ax.spines['bottom'].set_color('none')
			ax.spines['left'].set_color('none')            
		if debugTF:
			sys.stderr.write("===== DF:\n{}\n".format(df.tail()))
		#ax.plot(df.index,df[ticker])
		#df.plot(ax=ax)
		ax.plot(df.index,df.iloc[:,0])
		x_fmt = getKeyVal(optx,'x_fmt','') 
		yunit=['','1,000','Million','Billion','Trillion']
		ymax = df.iloc[:,0].max()
		ndg = max(int(math.log(ymax,10)/3),1)
		nadj = (ndg-1)*3
		ylabel = yunit[ndg-1] or ''
		if debugTF:
			sys.stderr.write("===== {} {} {} {}\n".format(ymax,ndg,nadj,yunit[ndg-1]))
		if ndg>1:
			ax.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))
		else:
			ax.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x)))
		ax.set_ylabel(ylabel)
		if trendTF is True:
			npar = getKeyVal(optx,'npar',6)
			trendName = df._get_numeric_data().columns[0]
			dy =  df[trendName]
			dg, dh = vertex_locator(dy,npar=npar,debugTF=debugTF)
			ax.plot(dg.index,dg.iloc[:,0])
			if debugTF:
				sys.stderr.write("===plot data:\n {}\n".format(dg.tail()))
		if len(dv)>0 and isinstance(dv,pd.DataFrame):
			axv = ax.twinx()
			kind2='area' if src!='fred' else 'line'
			color2='lightgray' if src!='fred' else 'gray'
			axv.grid(False, axis='both')            
			axv.spines['right'].set_color('none')
			axv.spines['top'].set_color('none')
			axv.spines['bottom'].set_color('none')
			axv.spines['left'].set_color('none') 
			axv.plot(dv.index,dv.iloc[:,0],color=color2,alpha=.4)
			y2max = dv.iloc[:,0].max()
			ndg = max(int(math.log(y2max,10)/3),1)
			nadj = (ndg-1)*3
			if ndg>1:
				axv.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))
			else:
				axv.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x)))
			ylabel2 = yunit[ndg-1] or ''
			axv.set_ylabel(ylabel2)
			if debugTF:
				sys.stderr.write("===plot data:\n {}\n".format(dv.tail()))
		if debugTF:
			sys.stderr.write("===plot data:\n {}\n".format(df.tail(20)))
		plt.subplots_adjust(left=0.1,bottom=bottom)
		plt.title(title,fontproperties=fontProp())
		if len(x_fmt)>0: # TBD, specify xticks format
			ax.xaxis.set_major_formatter(mdates.DateFormatter(x_fmt))
		ax.grid(linestyle='dotted',linewidth=0.5)
		if chartname is not None and len(chartname)>5:
			plt.savefig(chartpath, format=chartformat) #, bbox_inches='tight',dpi=1000)
		elif backend is not None and backend.lower()=='tkagg': 
			plt.show()
	except Exception as e:
		sys.stderr.write("**ERROR: {} of {} @{}\n".format(str(e),chartpath,"plot_daily_macro"))
		return '',None,None
	return(chartpath,ax,fig)

def plot_daily_mostactive(da,db=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	'''
	plot daily_peers by marketcap barh
	'''
	sys.stderr.write("===start:{}()\n".format("plot_daily_mostactive"))
	ticker = getKeyVal(optx,'ticker')
	if backend is not None:
		plt.switch_backend(backend)
	chartpath,chartname = gen_chartpath(**optx)
	sns.set(style="whitegrid")

	# define plot
	if fig is None:
		fig, ax = plt.subplots(figsize=figsize)
	ax.grid(False, axis='y')
	ax.spines['right'].set_color('none')
	ax.spines['top'].set_color('none')
	ax.spines['bottom'].set_color('none')
	ax.spines['left'].set_color('none')

	bottom=optx['bottom'] if 'bottom' in optx else 0.3

	#da.plot(ax=ax,kind=kind,position=position,width=width)

	#xmax = da['changePercent'].max()
	#ndg = max(int(math.log(xmax,10)/3),1)
	#nadj = (ndg-1)*3
	#yunit=['','1,000','Million','Billion','Trillion']
	#ylabel = yunit[ndg-1] or ''
	#ax.xaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))
	da = da.sort_values(by='changePercent')        
	tmp = ax.barh(y=da['ticker'], width=da['changePercent'])

	#- assign targeted ticker with distinct color
	pos=int(da[da['ticker']==ticker].index.values)
	ax.patches[pos].set_facecolor('#aa3333')        
    
	da['pos'] = range(0,len(da))         
	#pos=int(dh[dh.index==ticker]['pos'])
	#axes.patches[pos].set_facecolor('#aa3333')        
	for sbar in tmp:
		height = sbar.get_width()            
		ax.annotate('{}'.format(round(height,2)),xy=(height, sbar.get_y()),xytext=((3 if height>0 else -3),3),textcoords="offset points",ha=('left' if height>0 else 'right'), va='bottom')
	#plt.legend(frameon=False) 

	plt.xlim((da['changePercent'].min()*1.3 if da['changePercent'].min()<0 else 0) , right = da['changePercent'].max()*1.2)
        
	plt.title('changePercent')    
	#barchart = sns.barplot(x="changePercent", y='ticker', data=da,palette="Blues_d")
	#barchart.set(ylabel='ticker')#xlabel=ylabel)
	#pos=int(da[da['ticker']==ticker].index.values)
	#ax.patches[pos].set_facecolor('#aa3333')
	# plot features
	loc=optx['loc'] if 'loc' in optx else 9
	ncol=optx['ncol'] if 'ncol' in optx else 2
	fancybox=optx['fancybox'] if 'fancybox' in optx else True
	borderaxespad=optx['borderaxespad'] if 'borderaxespad' in optx else 0.
	#plt.legend(loc=loc, ncol=ncol, fancybox=fancybox, borderaxespad=borderaxespad)
	rotation=optx['rotation'] if 'rotation' in optx else '30'
	fontsize=optx['fontsize'] if 'fontsize' in optx else 12
	plt.xticks(rotation=rotation,fontsize=fontsize)
	plt.legend(frameon=False)
	if len(title)>0:
	    plt.title(title,fontproperties=fontProp(size=15))
	plt.subplots_adjust(left=0.1,bottom=bottom)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return chartpath,ax,fig

def plot_stock_performance(df,fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=True,**optx):
	sys.stderr.write("===start:{}()\n".format("plot_stock_performance"))
	if isinstance(df,str):
		df,dv = get_data(df,**optx)
	elif not isinstance(df,pd.DataFrame):
		df = pd.DataFrame(data=df)
	if len(df)<1:
		return '',None,None
	if backend is not None and len(backend)>1:
		plt.switch_backend(backend)
	bottom=optx['bottom'] if 'bottom' in optx else 0.2
	chartpath,chartname = gen_chartpath(**optx)
	trendTF = getKeyVal(optx,'trendTF',True)    
	x_fmt = getKeyVal(optx,'x_fmt','%H:%M')    
	max_num = df['close'].max()
	max_num_date = df[df['close'] == max_num].iloc[-1].name
	min_num = np.min(df['close'])
	min_num_date = df[df['close'] == min_num].iloc[-1].name
	curr_num = df['close'][-1]
	curr_num_date = df[df['close'] == curr_num].iloc[-1].name
	open_num = df['close'].iloc[0]    
	max_index = np.max(df.index)
	min_index = np.min(df.index)
	x_lim = min_index.replace(hour=16, minute =31 )
	xprice = getKeyVal(optx,'xprice',open_num)
	cpchg = getKeyVal(optx,'cpchg',round((curr_num/open_num-1)*100 ,2)) 
	max_ylim = np.max([max_num,xprice])
	min_ylim = np.min([min_num,xprice])   
	try:
		from _alan_pppscf import vertex_locator
		if trendTF is True:
			try:
				npar = getKeyVal(optx,'npar',6)
				trendName = df._get_numeric_data().columns[0]
				dy = df[trendName]
				dg, dh = vertex_locator(dy,npar=npar,debugTF=debugTF)
				dh = dg
			except Exception as e:
				sys.stderr.write("**ERROR:{} @ {}\n".format(str(e),"running vertex_locator()"))
				dh={}
			#print(dh)            
		else:
			dh={}
		if not isinstance(df,pd.DataFrame):
			df = pd.DataFrame(data=df)
		if fig is None:
			plt.rcParams.update(plt.rcParamsDefault)            
			sns.set_style("whitegrid",{'grid.linestyle':'--'})                 
			fig, ax=plt.subplots(figsize=figsize)
			fig.autofmt_xdate()
			ax.yaxis.set_major_formatter(StrMethodFormatter('{x:,.0f}'))
			plt.xlim(min_index-timedelta(minutes=5), x_lim+timedelta(minutes=5))
			plt.ylim(min_ylim-(max_ylim-min_ylim)/2, max_ylim+(max_ylim-min_ylim)/5)
			plt.subplots_adjust(bottom=0.2)
			#plt.xlabel('Time',fontsize=15)
			#plt.ylabel('price',fontsize=15)
			#hours = mdates.HourLocator(interval = 1)
			h_fmt = mdates.DateFormatter(x_fmt)
			#ax.xaxis.set_major_locator(hours)
			ax.xaxis.set_major_formatter(h_fmt)
			ax.grid(False, axis='x')                        
			ax.spines['right'].set_color('none')
			ax.spines['top'].set_color('none')
			ax.spines['bottom'].set_color('none')
			ax.spines['left'].set_color('none')
			plt.title(title,fontproperties=fontProp(size=15))
			plt.legend(frameon=False)            
		plt.plot(df.index, df['close'])
		plt.plot([min_index,x_lim],[xprice,xprice],'k--',lw=1)# plot open dash line
		plt.plot([curr_num_date,curr_num_date],[xprice,curr_num],'k--',lw=1)# plot day pchg vertical line
		#plt.annotate('{}%'.format(round((curr_num/open_num-1)*100,2)), xy=(curr_num_date,open_num+(curr_num-open_num)/2),xytext=(0,6),fontsize=10,annotation_clip=True, textcoords="offset points",ha=('left'), va='center') # plot pchg number        
		plt.scatter(max_num_date,max_num,s=20,color = 'g',zorder = 19)#plot max point
		plt.scatter(min_num_date,min_num,s=20,color = 'r',zorder = 18)#plot min point
		plt.scatter(curr_num_date,curr_num,s=40,zorder = 20)#plot curr point
		plt.annotate('{}'.format(round(max_num,2)), xy=(max_num_date,max_num),xytext=(0,20),fontsize=10,arrowprops=dict(arrowstyle='simple'),annotation_clip=True, textcoords="offset points",ha=('right'), va='bottom') # plot max point number
		plt.annotate('{}'.format(round(min_num,2)), xy=(min_num_date,min_num),xytext=(0,-20),fontsize=10,arrowprops=dict(arrowstyle='simple'),annotation_clip=True, textcoords="offset points",ha=('left'), va='top') # plot min point number
		plt.annotate('{} | {}%'.format(round(curr_num,2),round(cpchg,2)), xy=(curr_num_date,curr_num),xytext=(0,20),fontsize=10,arrowprops=dict(arrowstyle='simple'),annotation_clip=True, textcoords="offset points",ha=('left'), va='bottom') # plot curr point number
		if len(dh)>1:
			if dh.iloc[-1].values<dh.iloc[-2].values:
				ar_color = 'red'
			else:
				ar_color= 'green'        
			plt.annotate('',xy=(dh.iloc[-1].name,dh.iloc[-1]*1.00),xytext=(dh.iloc[-2].name,dh.iloc[-2]*1.00), arrowprops=dict(facecolor=ar_color, shrink=0.05),annotation_clip=True) # plot trend arrow
			plt.plot(dh.index, dh.values)
		if 'volume' in df:                       
			axv = ax.twinx()
			axv.grid(False, axis='both')
			axv.spines['right'].set_color('none')
			axv.spines['top'].set_color('none')
			axv.spines['bottom'].set_color('none')
			axv.spines['left'].set_color('none')
			axv.xaxis.set_major_formatter(h_fmt)
			axv.set_yticklabels([]) 
			axv.tick_params(right="off")      
			plt.ylim(0,df['volume'].max()*3.5)            
			#axv.plot(df['volume'].index, df['volume'], color='gray',alpha=0.8)
			axv.fill_between(df['volume'].index,df['volume'], alpha =0.8)             
		if debugTF:
			sys.stderr.write("===plot data:\n {}\n".format(df.tail()))
		plt.subplots_adjust(left=0.1,bottom=bottom)
		if chartname is not None and len(chartname)>5:
			plt.savefig(chartpath, format=chartformat) #, bbox_inches='tight',dpi=1000)
		elif backend is not None and backend.lower()=='tkagg': 
			plt.show()
	except Exception as e:
		sys.stderr.write("**ERROR: {} @{}\n".format(str(e),"plot_stock_performance"))
		return '',None,None
	return(chartpath,ax,fig)


def plot_daily_peers(da,db=[],fig=None,ax=None,figsize=(11,6),backend='Agg',chartformat='svg',debugTF=False,**optx):
	'''
	plot daily_peers by marketcap barh
	'''
	sys.stderr.write("===start:{}()\n".format("plot_daily_peers"))
	sys.stderr.write(" --da:\n{}\n".format(da))
	sys.stderr.write(" --db:\n{}\n".format(db))
	if not isinstance(da,pd.DataFrame) or da is None or len(da)<1:
		return {},{},{}   
	ticker = getKeyVal(optx,'ticker')
	title  = getKeyVal(optx,'title','MarketCap')
	sys.stderr.write(" --title: {}, ticker:{}\n".format(title,ticker))
	if backend is not None:
		plt.switch_backend(backend)
	chartpath,chartname = gen_chartpath(**optx)
	sns.set(style="whitegrid")

	# define plot
	if fig is None:
		fig, ax = plt.subplots(figsize=figsize)
	ax.grid(False, axis='y')
	ax.spines['right'].set_color('none')
	ax.spines['top'].set_color('none')
	ax.spines['bottom'].set_color('none')
	ax.spines['left'].set_color('none')

	bottom=optx['bottom'] if 'bottom' in optx else 0.3

	#da.plot(ax=ax,kind=kind,position=position,width=width)

	xmax = da['marketcap'].max()
	ndg = max(int(math.log(xmax,10)/3),1)
	nadj = (ndg-1)*3
	yunit=['','1,000','Million','Billion','Trillion']
	ylabel = yunit[ndg-1] or ''
	ax.xaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))

	barchart = sns.barplot(x="marketcap", y='peers', data=da,palette="Blues_d")
	barchart.set(xlabel=ylabel, ylabel='Peers')
	pos=int(da[da['peers']==ticker].index.values)
	ax.patches[pos].set_facecolor('#aa3333')
	# plot features
	loc=optx['loc'] if 'loc' in optx else 0
	ncol=optx['ncol'] if 'ncol' in optx else 0
	#if debugTF:
	sys.stderr.write(" --xmax,ndg,nadj,yuni: {} {} {} {}\n".format(xmax,ndg,nadj,yunit[ndg-1]))
	sys.stderr.write(" --optx:{}\n --loc:{},ncol:{}\n".format(optx,loc,ncol))
	fancybox=optx['fancybox'] if 'fancybox' in optx else True
	borderaxespad=optx['borderaxespad'] if 'borderaxespad' in optx else 0.
	plt.xlim(0,xmax*1.2)
	plt.title(title,fontproperties=fontProp(size=15))
	plt.legend(loc=loc, ncol=ncol, fancybox=fancybox, borderaxespad=borderaxespad)
	rotation=optx['rotation'] if 'rotation' in optx else '30'
	fontsize=optx['fontsize'] if 'fontsize' in optx else 12
	plt.xticks(rotation=rotation,fontsize=fontsize)
	plt.legend(frameon=False)
	plt.subplots_adjust(left=0.1,bottom=bottom)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return chartpath,ax,fig

def plot_peers_performance(df,fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	sns.set(style="whitegrid")
	df['pchg']=df['pchg']*100
	if 'ticker' in df and 'peRatio' in df:
		df = pd.DataFrame(df.sort_values(by='peRatio').set_index('ticker',drop=True))
	elif 'ticker' in df:
		df = pd.DataFrame(df.set_index('ticker',drop=True))
	ngrid = df.shape[1]
	kind = getKeyVal(optx,'kind','barh')
	pctTF = getKeyVal(optx,'pctTF',False)
	vtitle = getKeyVal(optx,'vtitle',["1-Day Price Change %","PE Ratio"])
	ticker = getKeyVal(optx,'ticker')    
	if pctTF is True:
		vfmt = getKeyVal(optx,'vfmt',["{:.2}%","{:.0f}"])
	else:
		vfmt = getKeyVal(optx,'vfmt',["{:.2%}","{:.0f}"])
	if len(df)<1:
		return '',None,None
	if backend is not None and len(backend)>1:
		plt.switch_backend(backend)
	bottom=optx['bottom'] if 'bottom' in optx else 0.3
	#style = getKeyVal(optx,'style','classic')
	#plt.style.use(style)
	chartpath,chartname = gen_chartpath(**optx)
	if ngrid>1: 
		if np.isnan(df.loc[ticker,'peRatio']):
			peTF = False
		else:
			peTF = True
	#print(peTF)
	if peTF is True:
		ncols = 2
		fig, axes = plt.subplots(nrows=1, ncols=ncols,figsize=figsize)
		for i in range(ncols):
			axes[i].spines['right'].set_color('none')
			axes[i].spines['top'].set_color('none')
			axes[i].spines['bottom'].set_color('none')
			axes[i].spines['left'].set_color('none')
			axes[i].set_title(vtitle[i])          
			axes[i].grid(False,axis= 'y')
			field = 'pchg' if i == 0 else 'peRatio'            
			dh = df.sort_values(by=field)
			plt.xlim((dh[field].min()*1.3 if dh[field].min()<0 else 0) , right = dh[field].max()*1.2)           
			tmp=axes[i].barh(y=dh.index.values, width=dh[field])
			dh['pos'] = range(0,len(dh))         
			pos=int(dh[dh.index==ticker]['pos'])
			axes[i].patches[pos].set_facecolor('#aa3333')      
			for sbar in tmp:
				height = sbar.get_width()            
				axes[i].annotate('{}'.format(round(height,2)),xy=(height, sbar.get_y()),xytext=((3 if height>0 else -3),3),textcoords="offset points",ha=('left' if height>0 else 'right'), va='bottom')
                                                                                             
	else:
		ncols = 1
		fig, axes = plt.subplots(nrows=1, ncols=ncols,figsize=figsize)
		axes.spines['right'].set_color('none')
		axes.spines['top'].set_color('none')
		axes.spines['bottom'].set_color('none')
		axes.spines['left'].set_color('none')
		axes.set_title(vtitle[0])    
		axes.grid(False, axis='y')
		dh = df.sort_values(by='pchg')        
		tmp = axes.barh(y=dh.index.values, width=dh['pchg'])
		dh['pos'] = range(0,len(dh))         
		pos=int(dh[dh.index==ticker]['pos'])
		axes.patches[pos].set_facecolor('#aa3333')        
		for sbar in tmp:
			height = sbar.get_width()            
			axes.annotate('{}'.format(round(height,2)),xy=(height, sbar.get_y()),xytext=((3 if height>0 else -3),3),textcoords="offset points",ha=('left' if height>0 else 'right'), va='bottom')
	plt.legend(frameon=False)       
	#axes = df.plot(kind=kind,legend=False,ax=plt.gca(),subplots=True,sharey=True)
	#for j in range(ngrid):
		#if vtitle[j] is not None and len(vtitle[j])>0:
			#ax0.set_title(vtitle[j])
		#for i, v in enumerate(df.iloc[:,j]):
			#ax1.text(v , i + .05, vfmt[j].format(v), fontweight='bold')
	plt.subplots_adjust(left=.14, bottom=bottom, right=.90, top=.90, wspace=.20, hspace=.50)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat) #, bbox_inches='tight',dpi=1000)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return(chartpath,axes,fig)

def plot_earnings_performance(da,db=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	'''
	plot daily_peers by marketcap barh
	'''
	ticker = getKeyVal(optx,'ticker')
	if backend is not None:
		plt.switch_backend(backend)
	chartpath,chartname = gen_chartpath(**optx)
	sns.set(style="whitegrid")

	# define plot
	if fig is None:
		fig, ax = plt.subplots(figsize=figsize)
	ax.grid(False, axis='x')
	ax.spines['right'].set_color('none')
	ax.spines['top'].set_color('none')
	ax.spines['bottom'].set_color('none')
	ax.spines['left'].set_color('none')

	bottom=optx['bottom'] if 'bottom' in optx else 0.3
	da['pbdate'] = da.index.astype(str)
	da = da.sort_index(ascending=True)
	#da.index = pd.to_datetime(da.index)    
	#plt.ylim(,da['actualEPS'].max()*1.2)
	plt.title('EPS')

	xmax = abs(da['actualEPS']).max()
	ndg = max(int(math.log(xmax,10)/3),1)
	nadj = (ndg-1)*3
	yunit=['','1,000','Million','Billion','Trillion']
	ylabel = yunit[ndg-1] or ''
	if debugTF:
		sys.stderr.write(" --xmax:{},ndg:{},nadj:{},yunit:{}\n".format(xmax,ndg,nadj,yunit[ndg-1]))
	ax.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))
	tmp = ax.bar(x=da['pbdate'].values, height=da['actualEPS'])
	#tmp = sns.barplot(x='pbdate',y='actualEPS',data=da,palette="Blues_d")
	#height = tmp.get_height()
	for rect in tmp:
		height = rect.get_height()
		ax.annotate('{}'.format(height),xy=(rect.get_x()+rect.get_width() / 2, height),xytext=(0, 3),textcoords="offset points",ha='center', va='bottom')    
	#sns.catplot('pbdate','actualEPS',data=da,kind='point')
	#barchart = sns.barplot(x='pbdate',y='actualEPS',data=da,palette="Blues_d")
	#barchart.set(xlabel=ylabel, ylabel='actualEPS')
	#for index, row in da.iterrows():
		#rint(row)        
		#barchart.text(row['pbdate'],row['actualEPS'], round(row['actualEPS'],2), color='black', ha="center")
	#sns.catplot('pbdate','actualEPS',data=da,kind='point')
	#print(da)
	pos=len(da)-1
	ax.patches[pos].set_facecolor('#aa3333')
	# plot features
	loc=optx['loc'] if 'loc' in optx else 9
	ncol=optx['ncol'] if 'ncol' in optx else 2
	fancybox=optx['fancybox'] if 'fancybox' in optx else True
	borderaxespad=optx['borderaxespad'] if 'borderaxespad' in optx else 0.
	plt.legend(loc=loc, ncol=ncol, fancybox=fancybox, borderaxespad=borderaxespad)
	rotation=optx['rotation'] if 'rotation' in optx else '30'
	fontsize=optx['fontsize'] if 'fontsize' in optx else 12
	plt.xticks(rotation=rotation,fontsize=fontsize)
	plt.legend(frameon=False)
	if len(title)>0:
		plt.title(title)
	plt.subplots_adjust(left=0.1,bottom=bottom)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return chartpath,ax,fig

def plot_ts(df,dv=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	if len(df)<1:
		return '',None,None
	sys.stderr.write("===start:{}()\n".format("plot_ts"))
	src = getKeyVal(optx,'src','yh')
	#if isinstance(df,str):
	if not isinstance(df,pd.DataFrame):
		ticker=df
		from _alan_calc import pullStockHistory as psd
		ranged = getKeyVal(optx,'ranged','1d')
		gap = getKeyVal(optx,'gap','1m')
		datax = psd(ticker,gap=gap,ranged=ranged,days=1500,src=src)
		datax.rename(columns={'close':ticker}, inplace=True)
		df = datax[[ticker]]
		if src=='fred':
			tk2=ticker+'_PCTCHG'
			dx2 = psd(tk2,src=src,days=1500)
			if len(dx2)>0:
				dv = datax[[ticker]]
				df = dx2[['close']]
		elif 'volume' in datax:
			dv = datax[['volume']]
	if 'volume' in df:
		dv = df[['volume']].copy()
		cs = list(df.columns)
		cs.pop(cs.index('volume'))
		df = df[cs]
	if backend is not None and len(backend)>1:
		plt.switch_backend(backend)
	trendTF = getKeyVal(optx,'trendTF',True)
	bottom=optx['bottom'] if 'bottom' in optx else 0.35
	chartpath,chartname = gen_chartpath(**optx)
	try:
		from _alan_pppscf import vertex_locator
		if not isinstance(df,pd.DataFrame):
			df = pd.DataFrame(data=df)
		if fig is None:
			sns.set_style("whitegrid",{'grid.linestyle':'--'})            
			fig, ax=plt.subplots(figsize=figsize)
			ax.grid(False, axis='x')
			ax.spines['right'].set_color('none')
			ax.spines['top'].set_color('none')
			ax.spines['bottom'].set_color('none')
			ax.spines['left'].set_color('none')            
		plt.plot(df.index,df.iloc[:,0])
		plt.title(title,fontproperties=fontProp(size=30))
		x_fmt = getKeyVal(optx,'x_fmt','') 
		yunit=['','1,000','Million','Billion','Trillion']
		ymax = abs(df.iloc[:,0]).max()
		ndg = max(int(math.log(ymax,10)/3),1)
		nadj = (ndg-1)*3
		ylabel = yunit[ndg-1] or ''
		if debugTF:
			sys.stderr.write("+++++ {} {} {} {}\n".format(ymax,ndg,nadj,yunit[ndg-1]))
		if ndg>1:
			ax.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))
		else:
			ax.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x)))
		ax.set_ylabel(ylabel)
		if trendTF is True:
			npar = getKeyVal(optx,'npar',6)
			trendName = df._get_numeric_data().columns[0]
			dy =  df[trendName]
			dg, dh = vertex_locator(dy,npar=npar,debugTF=debugTF)
			dg.plot(ax=ax,legend=False)
			if debugTF:
				sys.stderr.write("===plot data:\n {}\n".format(dg.tail()))
		if len(dv)>0 and isinstance(dv,pd.DataFrame):
			axv = ax.twinx()
			kind2='area' if src!='fred' else 'line'
			kind2 = getKeyVal(optx,'kind2',kind2)
			color2='lightgray' if src!='fred' else 'gray'
			if kind2 == 'bar':
				width2 = getKeyVal(optx,'width2',25)
				axv.bar(dv.index,dv.iloc[:,0],width=width2,color=color2,alpha=.4)
			else:
				axv = dv.plot(ax=axv,kind=kind2,color=color2,alpha=.4,legend=False)
			y2max = dv.iloc[:,0].max()
			ndg = max(int(math.log(y2max,10)/3),1)
			nadj = (ndg-1)*3
			if ndg>1:
				axv.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x/10**nadj)))
			else:
				axv.yaxis.set_major_formatter(FuncFormatter(lambda x,pos: '{:,g}'.format(x)))
			ylabel2 = yunit[ndg-1] or ''
			axv.set_ylabel(ylabel2)
		if debugTF:
			sys.stderr.write("===dv:\n {}\n".format(dv.tail(10)))
		if debugTF:
			sys.stderr.write("===df:\n {}\n".format(df.tail(10)))
		plt.subplots_adjust(left=0.1,bottom=bottom)
		if len(x_fmt)>0: # TBD, specify xticks format
			ax.xaxis.set_major_formatter(mdates.DateFormatter(x_fmt))
		ax.grid(linestyle='dotted',linewidth=0.5)
		if chartname is not None and len(chartname)>5:
			plt.savefig(chartpath, format=chartformat) #, bbox_inches='tight',dpi=1000)
		elif backend is not None and backend.lower()=='tkagg': 
			plt.show()
	except Exception as e:
		sys.stderr.write("**ERROR: {} @{}\n".format(str(e),"plot_ts"))
		sys.stderr.write("{}\n".format(locals()))
		return '',None,None
	return(chartpath,ax,fig)

def plot_barh2chart(df,fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	if 'ticker' in df and 'peRatio' in df:
		df = pd.DataFrame(df.dropna().sort_values(by='peRatio').set_index('ticker',drop=True))
	elif 'ticker' in df:
		df = pd.DataFrame(df.dropna().set_index('ticker',drop=True))
	ngrid = df.shape[1]
	kind = getKeyVal(optx,'kind','barh')
	pctTF = getKeyVal(optx,'pctTF',False)
	vtitle = getKeyVal(optx,'vtitle',["1-Day Price Change %","PE Ratio"])
	color = getKeyVal(optx,'color',['blue','red','green','gray','salmon','purple'])
	if pctTF is True:
		vfmt = getKeyVal(optx,'vfmt',["{:.2}%","{:.0f}"])
	else:
		vfmt = getKeyVal(optx,'vfmt',["{:.2%}","{:.0f}"])
	if len(df)<1:
		return '',None,None
	if backend is not None and len(backend)>1:
		plt.switch_backend(backend)
	bottom=optx['bottom'] if 'bottom' in optx else 0.35
	style = getKeyVal(optx,'style','classic')
	plt.style.use(style)
	chartpath,chartname = gen_chartpath(**optx)
	fig, axes = plt.subplots(ngrid,figsize=figsize,sharey=True)
	axes = df.plot(kind=kind,legend=False,ax=plt.gca(),color=color,subplots=True,sharey=True)
	for j in range(ngrid):
		if vtitle[j] is not None and len(vtitle[j])>0:
			axes[j].set_title(vtitle[j])
		for i, v in enumerate(df.iloc[:,j]):
			axes[j].text(v , i + .05, vfmt[j].format(v), fontweight='bold')
	#plt.gca().invert_yaxis() 
	plt.subplots_adjust(left=.1, bottom=bottom, right=.90, top=.90, wspace=.20, hspace=.50)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat) #, bbox_inches='tight',dpi=1000)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return(chartpath,axes,fig)

def plot_2yaxis(da,db=[],fig=None,ax=None,figsize=(11,6),backend='Agg',title='',chartformat='svg',debugTF=False,**optx):
	'''
	plot 2-YAxis via da and db data
	'''
	if backend is not None:
		plt.switch_backend(backend)
	chartpath,chartname = gen_chartpath(**optx)
	# define plot
	if fig is None:
		fig, ax = plt.subplots(figsize=figsize)
	# 1st plot
	bottom=optx['bottom'] if 'bottom' in optx else 0.35
	kind=optx['kind'] if 'kind' in optx else 'bar'
	color=optx['color'] if 'color' in optx else 'blue'
	width=optx['width'] if 'width' in optx else 0.2
	position=optx['position'] if 'position' in optx else 1.2
	da.plot(ax=ax,kind=kind,position=position,width=width)

	# 2nd plot
	if len(db)>0:
		ax2 = ax.twinx() 
		kind2=optx['kind2'] if 'kind2' in optx else kind
		color2=optx['color2'] if 'color2' in optx else 'red'
		width2=optx['width2'] if 'width2' in optx else width
		position2=optx['position2'] if 'position2' in optx else 0.3
		db.plot(ax=ax2,kind=kind2,color=color2,position=position2,width=width2)

	# plot features
	loc=optx['loc'] if 'loc' in optx else 9
	ncol=optx['ncol'] if 'ncol' in optx else 2
	fancybox=optx['fancybox'] if 'fancybox' in optx else True
	borderaxespad=optx['borderaxespad'] if 'borderaxespad' in optx else 0.
	plt.legend(loc=loc, ncol=ncol, fancybox=fancybox, borderaxespad=borderaxespad)
	rotation=optx['rotation'] if 'rotation' in optx else '30'
	fontsize=optx['fontsize'] if 'fontsize' in optx else 12
	plt.xticks(rotation=rotation,fontsize=fontsize)

	if len(title)>0:
		plt.title(title)
	plt.subplots_adjust(left=0.1,bottom=bottom)
	if chartname is not None and len(chartname)>5:
		plt.savefig(chartpath, format=chartformat)
	elif backend is not None and backend.lower()=='tkagg': 
		plt.show()
	return chartpath,ax,fig

def plot_templates(*args, **opts):
	funcname = getKeyVal(opts,'funcname','plot_ts')
	style = getKeyVal(opts,'style','classic')
	plt.style.use(style)
	if funcname in globals():
		funcArg = globals()[funcname]
	else:
		funcArg = plot_ts
	opts.pop('funcname',None)
	try:
		ret = funcArg(args[0],**opts)
	except Exception as e:
		sys.stderr.write("**ERROR:{} @{}\n".format(str(e),funcname))
		ret= str(e),None,None
	return ret

if __name__ == '__main__':
	from _alan_calc import pullStockHistory as psd
	#df = pd.read_csv("BTCUSD.dat",sep="|")
	#plot_barh2chart(df[['open','close']].tail(), chartname='',backend='tkAgg')
	#plot_ts('^GSPC', chartname='',backend='tkAgg',debugTF=True)
	#plot_ts('CPIAUCNS', chartname='',backend='tkAgg',debugTF=True,src='fred')
	#plot_2yaxis(df[['close']].iloc[:10], backend='tkAgg')
	from _alan_calc import pullStockHistory as psd
	df = psd('^GSPC',gap='1m',ranged='1d')
	plot_ts(df[['close','volume']], chartname='',backend='tkAgg',debugTF=True)
	#plot_intraday_headline(chartname='',backend='tkAgg',debugTF=True)



#!/usr/bin/env python3
'''
Update mapping_ticker_cik table (weekly cron every Saturday)
Usage of:
upd_mapping_ticker.py upd
OR
upd_mapping_ticker.py deact
OR
upd_mapping_ticker.py add PDD BILI
OR (DEPRECATED using 'add' option instead)
echo PDD |python3 -c "import sys;tkLst=sys.stdin.read().strip().split();from upd_mapping_ticker import add_new_ticker_cik as atc;atc(tkLst)"

Required programs:
  upd_mapping_ticker.py
  yh_chart.py 
  ticker_mapping.py
  theme_list.py
Required tables:
  PGDB::ara sp500_component, mapping_ticker_cik
  MDB::ara madmoney_hist

Last Mod., Fri Sep 17 15:24:36 EDT 2021
'''
import sys
from _alan_date import next_date
from _alan_calc import conn2mgdb, sqlQuery, conn2pgdb
from _alan_str import insert_mdb
from yh_chart import yh_quote_comparison as yqc
from ticker_mapping import ticker_mapping as tkmp

def add_new_ticker_cik(newLst=[],saveDB=True):
	if len(newLst)<1:
		return newLst
	newDf = tkmp(newLst)
	pgDB=conn2pgdb(dbname='ara')
	newDf.to_sql('ticker_mapping_temp',pgDB,schema='public',index=False,if_exists='replace')
	xqr="""
	DELETE FROM mapping_ticker_cik B USING ticker_mapping_temp C WHERE B.ticker = C.ticker;
	INSERT INTO mapping_ticker_cik
	SELECT cik,ticker,company,sic,sector,company_cn,sector_cn,1::int as act_code FROM (SELECT a.*,b.sector,b.sector_cn from ticker_mapping_temp a LEFT JOIN spdr_sector b ON a.sector_alias=b.sector_alias) as s
	"""
	pgDB.execute(xqr)
	return newLst

def add_mapping_ticker(lkbk=-7,mxcap=10**9,debugTF=False,saveDB=True,**optx):
	newLst=optx.pop('tkLst',[])
	if len(newLst)<1:
		sys.stderr.write("Empty List\n")
		return []
	else:
		sys.stderr.write("New List:{}\n".format(newLst))
	newLst = add_new_ticker_cik(newLst,saveDB)
	return newLst

def upd_mapping_ticker(lkbk=-7,mxcap=10**9,debugTF=False,saveDB=True,**optx):
	sys.stderr.write("===STARTING to update ticker list\n")
	#1. UPDATE sp500 / Dow30 list
	from theme_list import sp500_component
	sp500=sp500_component(saveDB=False)
	if len(sp500)<1:
		sp500=sqlQuery("SELECT * FROM sp500_component")

	sLst = sp500['ticker'].values

	#2. PULL MDB::ara:madmoney_hist
	dt6mo=int(next_date(months=-6,dformat='%Y%m01',dtTF=False))
	mdb=conn2mgdb(dbname='ara')
	mLst=mdb['madmoney_hist'].find({"pbdate":{"$gt":dt6mo}}).distinct("ticker")
	mLst = [x.replace('.','-') for x in mLst]

	#3. COMPILE a new list based on madmoney and SP500 new tickers
	tkCIK=sqlQuery("SELECT * FROM mapping_ticker_cik WHERE act_code=1 and ticker not similar to '%%(^|=)%%'")
	tLst = tkCIK['ticker'].values
	n1Lst = set(sLst)-set(tLst)
	sys.stderr.write("New SP500 List:{}\n".format(n1Lst))
	n2Lst = set(mLst)-set(tLst)
	sys.stderr.write("New madMoney List:{}\n".format(n2Lst))
	nLst = list( n1Lst.union(n2Lst) )

	#4. PULL only valid quotes within last 7 days and marketCap > 1B
	dt7d=int(next_date(days=lkbk,dformat='%Y%m%d',dtTF=False))
	df=yqc(nLst,tablename=None,screenerTF=False,dfTF=True)
	xqr='not(pbdate<{}|marketCap<{})'.format(dt7d,mxcap)
	dg = df.query(xqr)
	newLst=dg['ticker'].values
	if len(newLst)<1:
		sys.stderr.write("No Additions\n")
		return []
	else:
		sys.stderr.write("New Additional List:{}\n".format(newLst))

	#5. PULL new list 'newLst' with cik/sic/sector info
	newLst = add_new_ticker_cik(newLst,saveDB)
	return newLst

def deact_mapping_ticker(lkbk=-14,mxcap=500000000,debugTF=False,saveDB=True,**optx):
	'''
	De-activate tickers with invalid quotes within last 14 days or marketCap < 500MM
	'''
	# de-active stale tickers
	sys.stderr.write("===STARTING de-active stale tickers\n")
	pgDB=conn2pgdb(dbname='ara')
	tkCIK=sqlQuery("SELECT * FROM mapping_ticker_cik WHERE act_code=1 and ticker not similar to '%%(^|=)%%'")
	tLst = tkCIK['ticker'].values
	dt7d=int(next_date(days=lkbk,dformat='%Y%m%d',dtTF=False))
	df=yqc(tLst,tablename=None,screenerTF=False,dfTF=True)
	if len(df)<1:
		sys.stderr.write("No deactivate list\n")
		return(0)
	yqr='pbdate<{}|marketCap<{}'.format(dt7d,mxcap)
	dg = df.query(yqr)
	newLst=dg['ticker'].values
	if len(newLst)<1:
		sys.stderr.write("No deactivate list\n")
		return []
	else:
		sys.stderr.write("New deactivate List:{}\n".format(newLst))
	if saveDB:
		xqTmp="UPDATE mapping_ticker_cik set act_code=0 where ticker in {}"
		xqr = xqTmp.format(tuple(newLst))
		pgDB.execute(xqr)
	return newLst

def run_mappings():
	args=sys.argv[1:]
	if not args:
		sys.stderr.write("[OPTION] not given\n")
		sys.stderr.write("Usage of:\n{} [deact|upd|add] [TK1] [TK2] ...\n".format(sys.argv[0]))
		return []
	fx=args[0]
	tkLst=args[1:]
	d=dict(deact=deact_mapping_ticker,add=add_mapping_ticker,upd=upd_mapping_ticker)
	fn = d.pop(fx,None)
	if not fn:
		sys.stderr.write("[OPTION] not found\n")
		sys.stderr.write("Usage of:\n{} [deact|upd|add] [TK1] [TK2] ...\n".format(sys.argv[0]))
		return []
	return fn(tkLst=tkLst)

if __name__ == '__main__':
	run_mappings()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Program: lsi_daily.py
    Description: create ohlc commentary based on ohlc_hist and ohlc_pnl
    Input table required:
	prc_hist
	ohlc_pnl
	mapping_ticker_cik
	ohlc_latest_macd
	ohlc_ls_signal
	ohlc_pppscf
	ohlc_fcs 
	mapping_udf_comment_cn
	mapping_udf_comment_en
    Version: 0.66
    Output example: 
	蘋果08月16日每日快報：蘋果，代號A A P L，收盤價 $ 213.32。蘋果曾經在9天前出現符合烏雲蓋頂的反轉訊號，開始下降。目前，訊號所衍生的下降趨勢並未持續。 在過去12個月的價格囘側中，出現過23次烏雲蓋頂信號，獲利率為2%。 但是，MACD看漲的上昇訊號也在15天前出現，目前還在持續中並未改變，還在向上提升，可惜與烏雲蓋頂的方向並不一致。 此時趨勢呈現模糊不清的訊息，不建議採取任何進出行動。蘋果近期上升29個交易日, 總共為13%。 在過去14月裏其最大變幅在于07月06日和08月16日之間攀升13%。依據蘋果股價波動狀況，預估下一週價位有七成可能會在209到221元之間波動。上漲機率高達61%。
    Note:
	1. use MA(5) for confirm morning_star
	2. use MACD age and current level to confirm/conflict morning_star
    Function:
	def getdb_udfLst(category="stock",tbname="mapping_udf_comment",pgDB=None,lang="en"):
	def getdb_ticker(ticker,sqx,pgDB):
	def generate_daily_comment(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="en"):
	def generate_comment_pricing(f,ts=None,dotSign='點',usdSign='元',accmm='',prcn=2,lang="cn",mp3YN=False):
	def generate_comment_title(f,ts=None,dotSign='點',usdSign='元',accmm='',prcn=2,lang="cn",mp3YN=False):
	def assign_pppscf_ts(lang="en"):
	def generate_comment_pppscf(f,dotSign='.',prcn=0,usdSign='$ ',udfLst=None,lang="en",ts=None):
	def run_daily_comment(aX,enhanceX=None,tempS=None,dotSign='.',prcn=0,usdSign='$',lang="cn"):
	def getdb_ohlc_daily(ticker,startYmd,endYmd,sector,label,pgDB):
	def assign_title_ts(lang):
	def assign_pricing_ts(lang):
	def run_comment_pricing(ticker='AAPL',label='',pgDB=None,dotSign='點',prcn=2,usdSign='元',lang="cn",mp3YN=False,ts_pricing=None,ts_title=None):
	def run_comment_pppscf(ticker='AAPL',label='',pgDB=None,dotSign='.',prcn=0,usdSign='$',lang="cn",fp=None,ts=None):
	def generate_comment_fcst_en(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="en",mp3YN=False):
	def generate_comment_fcst_cn(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="cn",mp3YN=False):
	def generate_cmt(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="en",mp3YN=False,funcname=''):
	def run_comment_fcst(ticker='AAPL',label='',pgDB=None,dotSign='.',prcn=0,usdSign='',lang="cn",fp=None,ts=None,mp3YN=True):
	def save2mgdb_daily_comment(ret,dbM,tablename):
	def iteritem_daily_comment(j,tempS,tkX,mndt,mxdt,sector,label,daily_hdr,tablename,pgDB,mgDB,fp,wmode="replace",lang="cn",saveDB=True,mp3YN=False,dirname=None,debugTF=False):
	def assign_ts_ohlc(lang):
	def batch_daily_comment(tkLst=None,**kwargs):
	def opt_lsi_daily(argv,retParser=False):
    Last Mod., Fri Aug 17 14:47:09 EDT 2018
"""
from optparse import OptionParser
from _alan_str import *
from _alan_date import delta2dates
from _alan_calc import conn2pgdb,conn2mgdb,pqint
from bson import json_util
import pandas as pd
import os
import sys
if sys.version_info.major == 2:
	reload(sys)
	sys.setdefaultencoding('utf8')

def prn_dummy(du={}, fp=sys.stderr):
	""" print temp dict: du to file handle: fp
	"""
	if isinstance(fp,str) is True and len(fp)>0:
		fp = sys.stdout if fp == '-' else open(fp,'w')
	if isinstance(du,str) is True and len(du)>0:
			fp.write(du)
			fp.write("\n")
			return 1
	elif isinstance(du,dict) is False or len(du)<1:
		return None
	for k,v in  du.items():
		if isinstance(v,list) is True:
			for x in v:
				fp.write(x)
				fp.write("\n")
		elif isinstance(v,dict) is True:
			for j,x in v.items():
				fp.write(x)
				fp.write("\n")
		else:
			fp.write(v)
			fp.write("\n")

def assign_ts_title(lang):
	ts="{{label}}{{currDateWd}}每日快報：" if lang=="cn" else "{{label}} {{currDateWd}} Daily Alert: "
	return ts

def assign_ts_pricing(lang):
	ts="{{label}}，代號{{tickerWd}}，收盤價 {{priceWd}}。 " if lang=="cn" else "{{label}} ticker: {{tickerWd}}, closed at {{priceWd}}. "
	return ts

def assign_ts_pppscf(lang="en"):
	j2ts_cn="""{{label}}近期{{past_udf}}{{past_difday}}{{unitStr}}, 總共為{{past_act_chg_pct}}。 {%if act_chg_zscore > 1.96 %}
{# define syntax display #}在過去{{past_range}}裏其最大變幅在于{{latest_startYM}}和{{latest_endYM}}之間{{latest_udf}}{{latest_act_chg_pct}}。{% endif %}""" 
	j2ts_en="""{{label}} recently {{past_udf}} {{past_act_chg_pct}} for {{past_difday}} {{unitStr}}.
{# define syntax display #}{%if act_chg_zscore > 1.96 %}In the past {{past_range}}, the latest movement is between {{latest_startYM}} and {{latest_endYM}}, {{latest_udf}} {{latest_act_chg_pct}}.{% endif %}""" 
	return j2ts_en if lang=="en" else j2ts_cn

def assign_ts_ohlc(lang="en"):
	global tempStr_cn,macdStr_cn,macdEnhance_0_cn,macdEnhance_1_cn,tempStr_en,macdStr_en,macdEnhance_0_en,macdEnhance_1_en
	macdStr_en="""{{label}} {{pastAdv}} emerged the {{strategyName}} signal. So far, the {{comingTrendWd}} signal {{continueTrendWd}} and {{continueStatusClause}}. Backtest in the past {{monthWd}}，there are {{countWd}} of {{strategyName}} signals，{{trrAdv}} with total return at {{trrPct}}."""
	tempStr_en="""{{label}} {{pastAdv}} appeared a reversing signal {{strategyName}} {{pastDayWd}}, {{comingTrendWd}} started to emerge. Currently, the {{comingTrendWd}} signal {{continueTrendWd}} and {{continueStatusClause}}. Backtest in the past {{monthWd}}，there are {{countWd}} of {{strategyName}} signals，{{trrAdv}} with total return at {{trrPct}}. {{enhanceClause}} """
	macdEnhance_0_en="""{{prepositionWd}}，{{strategyName}} {{comingTrendWd}} also {{pastDayWd}} appeared，so far {{continueTrendWd}} {{continueStatusClause}}，its direction was not consistent with {{enhanceName}}. Those conflict singals do not suggest any action"""
	macdEnhance_1_en="""{{prepositionWd}}，{{strategyName}} {{comingTrendWd}} also {{pastDayWd}} appeared，so far {{continueTrendWd}} {{continueStatusClause}} and was consistent with {{enhanceName}}. Both singals enhanceed our suggestion."""
	macdStr_cn="""{{label}}{{pastAdv}}在{{pastDayWd}}出現符合{{strategyName}}的{{comingTrendWd}}訊號。目前，訊號所衍生的{{comingTrendWd}}趨勢{{continueTrendWd}}{{continueStatusClause}}。 在過去{{monthWd}}的價格囘側中，出現過{{countWd}}次{{strategyName}}信號，{{trrAdv}}獲利率為{{trrPct}}。 """
	tempStr_cn="""{{label}}{{pastAdv}}在{{pastDayWd}}出現符合{{strategyName}}的反轉訊號，開始{{comingTrendWd}}。目前，訊號所衍生的{{comingTrendWd}}趨勢{{continueTrendWd}}{{continueStatusClause}}。 在過去{{monthWd}}的價格囘側中，出現過{{countWd}}次{{strategyName}}信號，{{trrAdv}}獲利率為{{trrPct}}。 {{enhanceClause}}"""
	macdEnhance_0_cn="""{{prepositionWd}}，{{strategyName}}的{{comingTrendWd}}訊號也在{{pastDayWd}}出現，目前{{continueTrendWd}}{{continueStatusClause}}，可惜與{{enhanceName}}的方向並不一致。 此時趨勢呈現模糊不清的訊息，不建議採取任何進出行動。"""
	macdEnhance_1_cn="""{{prepositionWd}}，{{strategyName}}的{{comingTrendWd}}訊號也在{{pastDayWd}}出現，目前{{continueTrendWd}}{{continueStatusClause}}，與{{enhanceName}}的方向一致，進而發出更加明確的{{comingLookoutWd}}訊息。 在之前同一期間也有過{{countWd}}次{{strategyName}}，{{trrAdv}}獲利為{{trrPct}}{{trrAddClause}}。"""

	tempM={ "en":{"macd":macdStr_en,"temp":tempStr_en,"enhance":[macdEnhance_0_en,macdEnhance_1_en]},
		"cn":{"macd":macdStr_cn,"temp":tempStr_cn,"enhance":[macdEnhance_0_cn,macdEnhance_1_cn]}}
	#for lang in ["cn","en"]:
	daily_hdr=["ticker","name","price","trr","tsig","pnl_prd","curr_date","pbdate","lsc","curr_trend","curr_deriv","comment"]
	return (tempM[lang],daily_hdr)

def getdb_udfLst(category="stock",tbname="mapping_udf_comment",pgDB=None,lang="en"):
	try:
		if tbname and pgDB :
			tbname="{}_{}".format(tbname,lang) if lang != "en" else tbname
			sqx="SELECT * FROM {1} WHERE category={0!r} ORDER BY thd DESC"
			sqr=sqx.format(str(category),tbname)
			return pd.read_sql(sqr,pgDB)
		else:
			return None
	except Exception as e:
		pqint( "**ERROR @ getdb_udfLst():",str(e), file=sys.stderr)
		return None

def getdb_ticker(ticker,sqx,pgDB):
	try:
		gx=pd.read_sql(sqx.format(str(ticker)),pgDB)
	except:
		return None
	return gx

def generate_daily_comment(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="en"):
	""" stock ohlc strategy comment for each strategy
	    essential f keys: ticker,name,price,trr,tsig,pbdate,curr_date,lsc,curr_trend,curr_deriv
	"""
	if ts is None:
		return ''
	try:
		enhanceClause= f['enhanceClause']
	except:
		enhanceClause= ""
	if 'enhanceName' in f: # For "enhance" case ONLY
		enhanceName = pattern_cn(f['enhanceName'],lang=lang)
		ynCode=f['ynCode']
	else:
		enhanceName=None
		ynCode=1
	xnDay=delta2dates(f['curr_date'],f['pbdate']) # to calc days of previous stragety
	bsFlg = True if f['lsc'] == "long" else False
	currFlg = True if f['curr_trend']*f['amount'] >=0 else False # check if current situation still persists
	currChg = True if f['curr_deriv'] >=0  else False # if situation persists, is it enhancing or diminishing
	pastAdv = ""
	trrAdv = ""
	trrAddClause = ""
	if lang == "cn":
		currDateWd=ymd2md(str(f['curr_date']),lang=lang)
		buySellVerb = "買入" if bsFlg else "賣出"
		prepositionWd = rdWord("另外|同時|不但如此") if ynCode==1 else rdWord("然而|但是|不過|可是")
		pastTrendWd = "跌" if bsFlg else "漲" # reverse of long/short for beforeward harami/star signal
		comingLookoutWd = rdWord("看漲|看多") if bsFlg else rdWord("看跌|看空")
		comingTrendWd = rdWord("上昇|看多|向上提升") if bsFlg else rdWord("下降|看空|下滑")
		continueTrendWd = rdWord("還在持續中並未改變|依然持續|仍在持續中") if currFlg else rdWord("並沒有持續|並未持續")
		if currFlg is True:
			upWd=rdWord("增强|升高|上升|向上提升")
			dnWd=rdWord("減弱|下降|滑落|下滑")
			continueStatusClause = "，還在"+upWd if currChg else "，但已逐漸"+dnWd
		else:
			continueStatusClause = ""
		if xnDay < 1:
			pastDayWd = "昨日"
		elif xnDay < 2:
			pastDayWd = "前一天".format(xnDay)
		else:
			pastDayWd = "{}天前".format(xnDay)
			pastAdv = "曾經"
		monthWd = "{}個月".format(f['pnl_prd'])
		negWd = "而" if f['trr'] >-0.005 else rdWord("只是|但是|可惜|可是")
		trrAdv = "" if f['trr'] >-0.005 else rdWord("只是|但是|可惜|可是")
		trrAddClause = "" if f['trr'] >-0.005 else rdWord("，並非有效策略|，策略有待商榷") 
	else :
		prepositionWd = rdWord("In the mean time|In addition") if ynCode==1 else rdWord("However|Yet|Nevertheless")
		currDateWd=ymd2md(f['curr_date'],lang=lang)
		buySellVerb = "buy" if bsFlg else "sell"
		pastTrendWd = "downtrend" if bsFlg else "uptrend" # reverse of long/short for beforeward harami/star signal
		comingTrendWd = "uptrend" if bsFlg else "downtrend"
		continueTrendWd = "still persists" if currFlg else "starts to reverse"
		continueStatusClause = "continues to increase" if currChg else "starts to reduce"
		if xnDay < 1:
			pastDayWd = "just"
		elif xnDay < 2:
			pastDayWd = "yesterday".format(xnDay)
		else:
			pastDayWd = "{} days ago".format(xnDay)
		monthWd = "{}-month".format(f['pnl_prd'])
		negWd = "" if f['trr'] >-0.005 else "however" 
	label = f['label'] if 'label' in f else ''
	ticker = f['ticker']
	if label is None or str(label)=='':
		label=ticker
	price = roundUSD(f['price'],1)
	strategyName = f['name']
	strategyName = pattern_cn(strategyName,lang=lang)
	countWd = f['tsig']
	trrPct = roundPct(f['trr'],0,pct=1)
	if trrPct == '0%' or trrPct == '-0%':
		trrPct = "持平而已" if lang == "cn" else "flat"
	pricingComm = ""
	dux=locals()
	ret=jj_fmt(ts,dux)
	#ret=ts.format(**dux)
	return(ret)

def generate_comment_pricing(f,ts=None,dotSign='點',usdSign='元',accmm='',prcn=2,lang="cn",mp3YN=False):
	""" convert essential key words: ticker,label,and price to comment
	"{label}代號{tickerWd}，收盤價 {priceWd}。"
	"""
	tickerWd=" ".join(list(str(f['ticker'])))
	label=f['label']
	prX=f['price']
	currDateWd=ymd2md(str(f['curr_date']),lang=lang)
	rfcn=roundMP3 if mp3YN is True else roundUSD
	priceWd=rfcn(prX,prcn,usdSign=usdSign,dotSign=dotSign)
	if lang=="cn" :
		tsWd='{}代號{}，收盤價 {}。' if ts is None else ts
	else:
		tsWd='{} ticker: {} closed at {}. ' if ts is None else ts
	if mp3YN is False:
		priceWd='{} {:{}.{}f}'.format('$',prX,',',prcn)
	dux=locals()
	ret=jj_fmt(tsWd,dux)
	#ret = tsWd.format(**dux)
	return ret

def generate_comment_title(f,ts=None,dotSign='點',usdSign='元',accmm='',prcn=2,lang="cn",mp3YN=False):
	""" generate daily title
	"{label}{currDateWd}每日快報："
	"""
	tsWd="{label}{currDateWd}每日快報：" if ts is None else ts
	currDateWd=ymd2md(str(f['curr_date']),lang=lang)
	label=f['label']
	dux=locals()
	ret=jj_fmt(tsWd,dux)
	#ret = tsWd.format(**dux)
	return ret

def generate_comment_pppscf(f,dotSign='.',prcn=0,usdSign='$ ',udfLst=None,lang="en",ts=None):
	""" stock past performance comment
	"""
	if ts is None:
		ts = assign_ts_pppscf(lang=lang)
	fp=f.iloc[-1].to_dict()
	#for ky,va in fp.items():
	#	exec("{}=va".format(ky)) 
	unitStr=fq2unit_str(fp['freq'],lang)
	past_udf=udfStr(fp['act_chg_zscore'],udf=None,zs=0.0001,lang=lang)
	past_difday=fp['difday']
	past_act_chg_pct=roundPct(fp['act_chg'],n=0,pct=100,dotSign=dotSign,lang=lang)
	past_range=str(delta2dates(int(f.iloc[-1]['pbdate']),int(f.iloc[0]['pbdate']),fq='M')) + ("-month" if lang=="en" else "月")
	fx=f.query("sig_act_chg==1").iloc[0].to_dict()
	#for ky,va in fp.items():
	#	exec("{}=va".format(ky)) 
	act_chg_zscore=fx.pop('act_chg_zscore',0)
	act_chg=fx.pop('act_chg',0)
	xdate=fx.pop('xdate',0)
	pdate=fx.pop('pdate',0)
	latest_startYM=ymd2md(xdate,ymd='%Y-%m-%d',ym="%B %Y",lang=lang)
	latest_endYM=ymd2md(pdate,ymd='%Y-%m-%d',ym="%B %Y",lang=lang)
	latest_udf=udfStr(act_chg_zscore,udf=None,zs=0.0001,lang=lang)
	latest_act_chg_pct=roundPct(act_chg,n=0,pct=100,dotSign=dotSign,lang=lang)
	dux=locals()
	ret=jj_fmt(ts,dux)
	global gData
	gData=dux.copy()
	del gData['fp']
	del gData['fx']
	return ret

def run_daily_comment(aX,enhanceX=None,tempS=None,dotSign='.',prcn=0,usdSign='$',lang="cn"):
	if tempS is None:
		return None
	if aX is None and enhanceX is not None:
		aX = enhanceX
		enhanceX = None
		tempX=tempS["macd"]
	else:
		tempX=tempS["temp"]

	if enhanceX is not None:
		enhanceX['enhanceName']=str(aX['name'])
		ynCode = 1 if aX['lsc'] == enhanceX['lsc'] else 0 # check if both strategies for long/short are the same
		tempE=tempS['enhance'][ynCode]
		enhanceX['ynCode']=ynCode
		enhanceClause= generate_daily_comment(enhanceX,tempE,dotSign=dotSign,lang=lang)
		aX['enhanceClause']=enhanceClause
	else:
		aX['enhanceClause']=""

	ret=generate_daily_comment(aX,tempX,dotSign=dotSign,lang=lang)
	aX['nsig']=aX['tsig']
	aX['comment_ohlc']=ret
	return aX

def getdb_ohlc_daily(ticker,startYmd,endYmd,sector,label,pgDB):
	""" get info from ohlc_hist and ohlc_pnl for ohlc strategy info
	"""
	sq1="SELECT * FROM ohlc_latest_macd WHERE ticker = {0!r} order by pbdate DESC"
	sq2="SELECT * FROM ohlc_ls_signal WHERE ticker = {0!r} order by pbdate DESC"
	#sq1="SELECT ticker,pbdate,price,ma5,ma20,price-ma5 as ma5trend,signal_value_macd,signal_buysell_macd FROM ohlc_hist WHERE ticker = {0!r} and pbdate >= (SELECT max(pbdate) FROM ohlc_hist WHERE ticker = {0!r} and signal_buysell_macd <> 0 ) order by pbdate DESC"
	#sq2="select distinct * from ohlc_pnl where ticker={0!r} and sig=1 and lsc<>'combo' and name similar to '%%(star|harami|MACD)' order by name,lsc,pbdate DESC"
	g1=pd.read_sql(sq1.format(ticker),pgDB)
	g2=pd.read_sql(sq2.format(ticker),pgDB).loc[:,["ticker","name","pbdate","price","lsc","amount","cum_amounts","sig","tsig","ttr_hold","trr"]]
	pnl_prd=delta2dates(endYmd,startYmd,fq="M")
	pbdate=endYmd
	try:
		m1=g2.loc[~g2['name'].str.contains("MACD")].reset_index(drop=True).sort_values(by=['pbdate'],ascending=False).iloc[0]
		m1['curr_trend'] = m1['curr_deriv'] = 0
		if len(g1)>0:
			m1['curr_trend']=g1['ma5trend'][0] 
		if len(g1)>1:
			m1['curr_deriv'] = 1 if abs(g1['ma5trend'][0])>=abs(g1['ma5trend'][1]) else -1
		m1['curr_date']=pbdate
		m1['pnl_prd']=pnl_prd
		m1['sector']=sector
		m1['label']=label
	except:
		m1=None
	try:
		m2=g2.loc[g2['name'].str.contains("MACD")].reset_index(drop=True).sort_values(by=['pbdate'],ascending=False).iloc[0]
		m2['curr_trend'] = m2['curr_deriv'] = 0
		if len(g1)>0:
			m2['curr_trend'] = g1['signal_value_macd'][0]
		if len(g1)>1:
			m2['curr_deriv'] = 1 if abs(g1['signal_value_macd'][0])>=abs(g1['signal_value_macd'][1]) else -1
		m2['curr_date']=pbdate
		m2['pnl_prd']=pnl_prd
		m2['sector']=sector
		m2['label']=label
	except:
		m2=None
	return(pbdate,m1,m2)

def run_comment_pricing(ticker='AAPL',label='',pgDB=None,dotSign='點',prcn=2,usdSign='元',lang="cn",mp3YN=False,ts_pricing=None,ts_title=None):
	if ts_title is None:
		ts_title=assign_ts_title(lang)
	if ts_pricing is None:
		ts_pricing=assign_ts_pricing(lang)
	sqx="SELECT name as ticker, pbdate as curr_date,close as price,volume FROM prc_hist WHERE name={0!r} ORDER BY pbdate DESC limit 1"
	fp=getdb_ticker(ticker,sqx,pgDB).iloc[0]
	if 'label' not in fp:
		fp['label']=label
	try:
		comment_pricing = generate_comment_pricing(fp,ts=ts_pricing,dotSign=dotSign,prcn=prcn,usdSign=usdSign,lang=lang,mp3YN=mp3YN)
		comment_title = generate_comment_title(fp,ts=ts_title,dotSign=dotSign,prcn=prcn,usdSign=usdSign,lang=lang,mp3YN=mp3YN)
		pqint( "comment_pricing:",comment_pricing, file=sys.stderr)
		pqint( "comment_title:",comment_title, file=sys.stderr)
	except Exception as e:
		pqint( "**ERROR @ run_comment_pricing():",str(e), file=sys.stderr)
		return None
	return (comment_pricing,comment_title)

def run_comment_pppscf(ticker='AAPL',label='',pgDB=None,dotSign='.',prcn=0,usdSign='$',lang="cn",fp=None,ts=None):
	if fp is None:
		sqx="SELECT * FROM ohlc_pppscf WHERE ticker={0!r} ORDER BY pbdate"
		fp=getdb_ticker(ticker,sqx,pgDB)
	if 'label' not in fp:
		fp['label']=label
	if 'freq' not in fp:
		fp['freq']='D'
	if 'category' not in fp:
		fp['category']='stock'
	category=fp['category'].iloc[-1]

	#====Performance Comment====================
	udfLst=getdb_udfLst(category,"mapping_udf_comment",pgDB=pgDB,lang=lang)
	try:
		ret = generate_comment_pppscf(fp,dotSign=dotSign,prcn=prcn,usdSign=usdSign,udfLst=udfLst,lang=lang,ts=ts)
	except Exception as e:
		pqint( "**ERROR @ run_comment_pppscf():",str(e), file=sys.stderr)
		#pqint( fp, file=sys.stderr)
		return None
	return ret

def generate_comment_fcst_en(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="en",mp3YN=False):
	""" stock weekly forecast comment
		required f.fields: low_bound,up_bound,prc_cur,prc_x1w,prc_fcs,sigma,ticker,label,
		required ts fields: label,lowerWd,upperWd,posPbWd
	"""
	if ts is None:
		ts="""{label} closed {xdChgWd} at {price}, {xwChgWd} for the week. The {xwChgWd} movement is {cmpWd} the historical volatility of {sigmaWd}. Our {label} forecast for the next week is {nwTrendWd} with a probability of {posPbWd} and {plevelWd} chance of closing {rangeWd}."""
	#for (ky,va) in f.items(): 
	#	exec("{}=va".format(ky))
	low_bound,up_bound,prc_cur,prc_x1w,prc_fcs,xchg1d =\
		f['low_bound'],f['up_bound'],f['prc_cur'],f['prc_x1w'],f['prc_fcs'],f['xchg1d']
	prc_chg,rrate_sigma,plevel,pos_pb = f['prc_chg'],f['rrate_sigma'],f['plevel'],f['pos_pb']
	category = f['category']
	prc_dif=prc_cur-prc_x1w
	rfcn=roundMP3 if mp3YN is True else roundUSD
	rfcnPct=roundPctMP3 if mp3YN is True else roundPct
	unitStr=fq2unit_str(f['freq'],lang) if 'freq' in f else "週"
	xdTrendWd=udfStr(xchg1d,["up","down","unchanged"],0.005)
	xdChgWd=rfcn(xchg1d,2,usdSign=usdSign,dotSign=dotSign)
	xdayWd=rdWord("the previous trading day|previously|the last trading day")
	movingWd=rdWord("movement|change|difference")
	if xdTrendWd == "unchanged":
		xdChgWd = ''
	xwTrendWd=udfStr(prc_chg,["an increase","a decrease","unchanged"],0.0005)
	xwChgWd=rfcnPct(prc_chg,n=1,pct=100,dotSign=dotSign)
	xwDifWd=rfcn(prc_dif,2,usdSign=usdSign,dotSign=dotSign) 
	xwBpsWd=rfcn(prc_dif*100,0,usdSign=usdSign,dotSign=dotSign) 
	if xwTrendWd == "unchanged":
		xwChgWd = ''
	cmpWd=udfStr(abs(prc_chg)-rrate_sigma,["more than","less than","about the same as"],0.005)
	plevelWd=roundPct(plevel,n=0,pct=100)
	sigmaWd=rfcnPct(rrate_sigma,n=1,pct=100,dotSign=dotSign)
	posPbWd=roundPct(pos_pb,n=0,pct=100)
	nwTrendWd = "up"

	xt1wZ=prc_chg/rrate_sigma*100
	rt1w=f['prc_fcs']/f['prc_cur']
	rt1wZ=rt1w/f['sigma']*100
	pastTrendWd = performance_verb(xt1wZ,udfLst)
	comingTrendWd = performance_verb(rt1wZ,udfLst)

	vntDateWd=ymd2md(str(vntdate),ym="%B %d",lang=lang) if 'vntdate' in f else ""
	label = f['label'] if 'label' in f else ''
	ticker = f['ticker']
	category = f['category'] if 'category' in f else 'stock'
	unitSign='元' if lang=='cn' and category=='stock' else ''
	price = rfcn(f['prc_cur'],prcn,usdSign=usdSign,dotSign=dotSign)
	pn=0 if low_bound>99.4 else prcn
	lowerWd = rfcn(f['low_bound'],pn,usdSign=usdSign,dotSign=dotSign)
	upperWd = rfcn(f['up_bound'],pn,usdSign=usdSign,dotSign=dotSign)
	
	if label is None or str(label)=='':
		label=ticker
	rangeWd=btwStr(lowerWd,upperWd,["between","and"])

	dux=locals()
	ret=jj_fmt(ts,dux)
	for x,y in zip(["xdChgWd","xwChgWd","xwDifWd","xwBpsWd","sigmaWd","price","lowerWd","upperWd"],[xdChgWd,xwChgWd,xwDifWd,xwBpsWd,sigmaWd,price,lowerWd,upperWd]):
		dux.update({x:num2MP3(y,lang=lang)})
	ret_mp3=jj_fmt(ts,dux)
	global gData
	gData=dux.copy()
	del gData['f']
	return(ret,ret_mp3)

def generate_comment_fcst_cn(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="cn",mp3YN=False):
	""" stock weekly forecast comment
		required f.fields: low_bound,up_bound,prc_cur,prc_x1w,prc_fcs,sigma,ticker,label,
		required ts fields: label,lowerWd,upperWd,posPbWd
	"""
	freq="D"
	#for (ky,va) in f.items(): 
	#	exec("{}=va".format(ky))
	low_bound,up_bound,prc_cur,prc_x1w,prc_fcs,xchg1d =\
		f['low_bound'],f['up_bound'],f['prc_cur'],f['prc_x1w'],f['prc_fcs'],f['xchg1d']
	prc_chg,rrate_sigma,plevel,pos_pb = f['prc_chg'],f['rrate_sigma'],f['plevel'],f['pos_pb']
	category = f['category']
	prc_dif=prc_cur-prc_x1w
	rfcn=roundMP3 if mp3YN is True else roundUSD
	rfcnPct=roundPctMP3 if mp3YN is True else roundPct
	unitStr=fq2unit_str(f['freq'],lang) if 'freq' in f else "週"
	xdTrendWd=udfStr(xchg1d,None,0.005,lang=lang)
	xdChgWd=rfcn(xchg1d,2,usdSign=usdSign,dotSign=dotSign,lang=lang)
	xdayWd=rdWord("前一個交易日|前一天") if freq=="D" else rdWord("上一|前一")+unitStr
	movingWd=rdWord("變化|改變")
	if abs(xchg1d) <= 0.005:
		xdChgWd = ''
	xwTrendWd=udfStr(prc_chg,None,0.005,lang=lang)
	xwChgWd=rfcnPct(prc_chg,n=1,pct=100,dotSign=dotSign,lang=lang)
	xwDifWd=rfcn(prc_dif,2,usdSign=usdSign,dotSign=dotSign) 
	xwBpsWd=rfcn(prc_dif*100,0,usdSign=usdSign,dotSign=dotSign) 
	if abs(prc_chg) <= 0.0005:
		xwChgWd = ''
	cmpWd=udfStr(abs(prc_chg)-rrate_sigma,["大於","小於","相當於"],0.005)
	plevelWd=roundPct(plevel,n=0,pct=100)
	sigmaWd=rfcnPct(rrate_sigma,n=1,pct=100,dotSign=dotSign,lang=lang)
	posPbWd=roundPct(pos_pb,n=0,pct=100)
	nwTrendWd = "上升"

	xt1wZ=prc_chg/rrate_sigma*(1 if category=='stock' else 1)
	rt1w=f['prc_fcs']/f['prc_cur'] - 1
	rt1wZ=rt1w/f['sigma']*(1 if category=='stock' else 1)
	#xsign = 1 if ticker[-2:] == "US" else -1
	xsign = -1 # cn version always use EURO direction
	if category == 'currency':
		pastTrendWd = performance_verb(xt1wZ*xsign,udfLst)
		comingTrendWd = performance_verb(rt1wZ*xsign,udfLst)
	else:
		pastTrendWd = performance_verb(xt1wZ,udfLst)
		comingTrendWd = performance_verb(rt1wZ,udfLst)

	vntDateWd=ymd2md(str(vntdate),ym="%B %d",lang=lang) if 'vntdate' in f else ""
	label = f['label'] if 'label' in f else ''
	ticker = f['ticker']
	category = f['category'] if 'category' in f else 'stock'
	unitSign='元' if lang=='cn' and category=='stock' else ''
	price = rfcn(f['prc_cur'],prcn,usdSign=usdSign,dotSign=dotSign)
	pn=0 if low_bound>99.4 else prcn
	lowerWd = rfcn(f['low_bound'],pn,usdSign=usdSign,dotSign=dotSign)
	upperWd = rfcn(f['up_bound'],pn,usdSign=usdSign,dotSign=dotSign)
	
	if label is None or str(label)=='':
		label=ticker
	if 'pos_pb' in f:
		pp=f['pos_pb']*100.
		ppAt="大約為" if pp>=49.5 else "只達到"
		ppAt="高達" if pp>=59.5 else ppAt
		ppUp="貶值" if xsign==1 else "升值"
		ppUp="上漲" if category!="currency" else ppUp
		posPbWd="{}機率{}{:.0f}%。".format(ppUp,ppAt,pp)
	if lowerWd == upperWd:
		rangeWd = "變化不大，只會在{}{}內上下稍稍波動".format(lowerWd,unitSign) 
	else:
		rangeWd = "會在{}到{}{}之間波動".format(lowerWd,upperWd,unitSign) 
	dux=locals()
	ret=jj_fmt(ts,dux)
	for x,y in zip(["xdChgWd","xwChgWd","xwDifWd","xwBpsWd","sigmaWd","price","lowerWd","upperWd"],[xdChgWd,xwChgWd,xwDifWd,xwBpsWd,sigmaWd,price,lowerWd,upperWd]):
		dux.update({x:num2MP3(y,lang=lang)})
	ret_mp3=jj_fmt(ts,dux)
	global gData
	gData=dux.copy()
	del gData['f']
	return(ret,ret_mp3)

def generate_cmt(f,ts=None,dotSign='.',prcn=0,usdSign='$',udfLst=None,lang="en",mp3YN=False,funcname=''):
	try:
		funcN="{}_{}".format(funcname,lang)
		funcArg=globals()[funcN]
		xcmt=funcArg(f,ts=ts,dotSign=dotSign,prcn=prcn,usdSign=usdSign,udfLst=udfLst,mp3YN=mp3YN)
	except Exception as e:
		pqint( "**ERROR @ {}():{}".format("generate_cmt",str(e)), file=sys.stderr)
		xcmt=''
	return xcmt

def run_comment_fcst(ticker='AAPL',label='',pgDB=None,dotSign='.',prcn=0,usdSign='',lang="cn",fp=None,ts=None,mp3YN=True):
	if fp is None:
		#sqx="SELECT * FROM ara_weekly_forecast WHERE ticker={0!r} and freq='W'"
		sqx="SELECT * FROM ohlc_fcs WHERE ticker={0!r} and freq='W'"
		try:
			fp=getdb_ticker(ticker,sqx,pgDB).iloc[0]
		except:
			pqint( "No forecast available!", file=sys.stderr)
			return ''
	if 'label' not in fp:
		fp['label']=label
	if 'category' not in fp:
		fp['category']='stock'
	category=fp['category']
	
	if ts is None and lang=="cn":
		#ts="""{label}在前一週{pastTrendWd}，目前收盤價{price}元，我們預期在下一週將{comingTrendWd}，介于{lowerWd}至{upperWd}元之間。"""
		if category=='stock':
			ts="""依據{label}股價波動狀況，預估下一週價位有七成可能{rangeWd}。{posPbWd}"""
		else:
			ts="""依據{label}波動狀況，預估下一{unitStr}有七成可能{rangeWd}。{posPbWd}"""

	#====Performance Comment====================
	udfLst=getdb_udfLst(category,"mapping_udf_comment",pgDB=pgDB,lang=lang)
	ret,rmp3 = generate_cmt(fp,ts=ts,dotSign=dotSign,prcn=prcn,usdSign=usdSign,udfLst=udfLst,lang=lang,funcname="generate_comment_fcst",mp3YN=mp3YN)
	return ret

def save2mgdb_daily_comment(ret,dbM,tablename):
	if dbM is None:
		return None
	pqint( ret, file=sys.stderr)
	mobj=json_util.loads(ret.to_json())
	dbM[tablename].delete_one({"ticker":mobj["ticker"],"curr_date":mobj["curr_date"]})
	dbM[tablename].insert_one(mobj)
	return len(mobj)

def iteritem_daily_comment(j,tempS,tkX,mndt,mxdt,sector,label,daily_hdr,tablename,pgDB,mgDB,fp,wmode="replace",lang="cn",saveDB=True,mp3YN=False,dirname=None,debugTF=False):
	
	if mp3YN is True:
		dotSign = "點" if lang== "cn" else "point"
	else:
		dotSign = "."
			
	pqint("------",j,tkX,mndt,mxdt,sector,label, file=sys.stderr)
	try:
		(comment_pricing,comment_title)=run_comment_pricing(tkX,label,pgDB,dotSign=dotSign,lang=lang,mp3YN=mp3YN)
	except Exception as e:
		pqint( "**ERROR {}. {} @ {}: {}".format(j,tkX,"run_comment_pricing",str(e)), file=sys.stderr)
		comment_pricing=''
		comment_title=''
	try:
		comment_pppscf=run_comment_pppscf(tkX,label,pgDB,dotSign=dotSign,lang=lang)
		pqint( "{0}:{1}".format("comment_pppscf",comment_pppscf) , file=sys.stderr)
	except Exception as e:
		pqint( "**ERROR {}. {} @ {}: {}".format(j,tkX,"run_comment_pppscf",str(e)), file=sys.stderr)
		comment_pppscf=''
	try:
		comment_fcst=run_comment_fcst(tkX,label,pgDB,dotSign=dotSign,lang=lang,mp3YN=mp3YN)
		pqint( "{0}:{1}".format("comment_fcst",comment_fcst) , file=sys.stderr)
	except Exception as e:
		pqint( "**ERROR {}. {} @ {}: {}".format(j,tkX,"run_comment_fcst",str(e)), file=sys.stderr)
		comment_fcst=''
	try:
		(pbdate,m1,m2)=getdb_ohlc_daily(str(tkX),mndt,mxdt,sector,label,pgDB)
		aX=run_daily_comment(m1,m2,tempS,dotSign=dotSign,lang=lang)
		pqint( "{0}:{1}".format("comment_ohlc",aX['comment_ohlc']) , file=sys.stderr)
		# pqint(aX, file=sys.stderr)
	except Exception as e:
		pqint( "**ERROR {}. {} @ {}: {}".format(j,tkX,"run_daily_comment",str(e)), file=sys.stderr)
		return None

	aX['comment_pppscf']=comment_pppscf
	aX['comment_fcst']=comment_fcst
	aX['comment_pricing']=comment_pricing
	aX['comment_title']=comment_title
	comment='{comment_title}{comment_pricing}{comment_ohlc}{comment_pppscf}{comment_fcst}'.format(**aX)
	aX['comment']=comment
	pqint( "{0}:\n{1}".format("COMMENT",comment), file=sys.stderr)
	pqint( "{}".format(comment), file=sys.stdout)
	if debugTF:
		pqint( aX, file=sys.stderr)

	#= re-assign aX to tmpx and save to DB:tablename
	daily_hdr= daily_hdr + ["comment_title","comment_pppscf","comment_fcst","comment_pricing","comment_ohlc"]
	tmpx=aX[daily_hdr]

	if saveDB is True:
		#- save tmpx to database pgDB
		tmpx.to_frame().T.to_sql(tablename,pgDB,index=False,schema="public",if_exists=wmode)
		#- save tmpx to database mgDB
		save2mgdb_daily_comment(tmpx,mgDB,tablename)
		if debugTF is True:
			pqint( "save to [{}]".format(tablename), file=sys.stderr)
	else:
		if debugTF is True:
			pqint( "no save to table:[{}]".format(tablename), file=sys.stderr)

	#- write comment to file fdaily
	if dirname is not None and os.path.exists(dirname) is True:
		fdailyname="{}/daily_{}_{}_{}.txt".format(dirname,lang,aX['ticker'],aX['curr_date'])
		fdaily=open(fdailyname,"w")
		fdaily.write(aX['comment'])
		fdaily.close()
	#- write tmpx to file fp
	xstr="|".join(["{}"]*len(daily_hdr))+"\n"
	if fp is None:
		return tmpx
	if(j==0):
		fp.write(xstr.format(*daily_hdr))
	fp.write(xstr.format(*tmpx) )
	return tmpx

def batch_daily_comment(tkLst=None,lang='cn',wmode='replace',dirname=None,debugTF=False,hostname='localhost',tablename=None,filename=None,mp3YN=False,saveDB=True,dbname='ara',**kwargs):
	pgDB=conn2pgdb(dbname=dbname,hostname=hostname)
	mgDB=conn2mgdb(dbname=dbname.replace('.','_'),hostname=hostname)
	#====OHLC Comment===========================
	pqint( "====OHLC Comment===========================", file=sys.stderr)
	if tablename is None:
		tablename="{}_{}".format("ohlc_daily_comment",lang)
		if mp3YN is True:
			tablename="{}_{}".format(tablename,"mp3")
	else:
		tablename=tablename

	# Initial assignments
	(tempS,daily_hdr)=assign_ts_ohlc(lang=lang)
	if filename is None:
		fp = None
	elif filename == "stdout" or tablename == "-":
		fp = sys.stdout
	elif filename == "stderr":
		fp = sys.stderr
	else:
		fp = open(filename,"w") 

	# Get batch list
	sqr="SELECT s.*,m.sector,m.company{} as company FROM mapping_ticker_cik m right join (select ticker,min(pbdate) as mndt,max(pbdate) as mxdt from ohlc_pnl group by ticker) as s ON m.ticker=s.ticker order by s.ticker".format("" if lang=="en" else "_"+lang)
	dataM=pd.read_sql(sqr,pgDB)
	#dataM=pd.read_sql(sqr,pgDB)[9:10]
	if tkLst is not None:
		dx = dataM[dataM['ticker'].isin(tkLst)]
		if dx.empty is True:
			pqint("{} not found!".format(tkLst), file=sys.stderr)
			return None	
		else:
			dataM = dx.reset_index(drop=True)

	pqint("START batch_daily_comment-------------------------------------------", file=sys.stderr)
	for j, tRw in dataM.iterrows():
		(tkX,mndt,mxdt,sector,label)=tRw.values
		iteritem_daily_comment(j,tempS,tkX,mndt,mxdt,sector,label,daily_hdr,tablename,pgDB,mgDB,fp,wmode=wmode,lang=lang,saveDB=saveDB,mp3YN=mp3YN,dirname=dirname,debugTF=debugTF)
		wmode='append'
	#fp.close()
	pqint("END batch_daily_comment-------------------------------------------", file=sys.stderr)

def opt_lsi_daily(argv,retParser=False):
	""" command-line options initial setup
	    Arguments:
		argv:   list arguments, usually passed from sys.argv
		retParser:      OptionParser class return flag, default to False
	    Return: (options, args) tuple if retParser is False else OptionParser class
	"""
	parser = OptionParser(usage="usage: %prog [option] SYMBOL1 ...", version="%prog 0.66",
		description="Create ohlc commentary based on ohlc_hist and ohlc_pnl")
	parser.add_option("-d","--database",action="store",dest="dbname",default="ara",
		help="database name (default: eSTAR_2)")
	parser.add_option("","--host",action="store",dest="hostname",default="localhost",
		help="db host name (default: localhost)")
	parser.add_option("-t","--table",action="store",dest="tablename",
		help="db tablename (default: None)")
	parser.add_option("","--file",action="store",dest="filename",
		help="db filename (default: None)")
	parser.add_option("","--dirname",action="store",dest="dirname",
		help="db dirname to save comment of each ticker (default: None)")
	parser.add_option("-w","--wmode",action="store",dest="wmode",default="replace",
		help="db table write-mode [replace|append|fail] (default: replace)")
	parser.add_option("-l","--lang",action="store",dest="lang",default="cn",
		help="db language mode [cn|en] (default: cn)")
	parser.add_option("","--no_database_save",action="store_false",dest="saveDB",default=True,
		help="no save to database (default: save to database)")
	parser.add_option("","--use_mp3",action="store_true",dest="mp3YN",default=False,
		help="comment use mp3 style")
	parser.add_option("","--debug",action="store_true",dest="debugTF",default=False,
		help="debugging (default: False)")
	(options, args) = parser.parse_args(argv[1:])
	if retParser is True:
		return parser
	return (vars(options), args)

if __name__ == "__main__" :
	(opts, args)=opt_lsi_daily(sys.argv)
	tkLst=None
	if len(args)>0:
		if args[0]=='-':
			tkLst = sys.stdin.read().strip().split() 
		else:
			tkLst=args 
			opts['saveDB']=False
	batch_daily_comment(tkLst,**opts)



__doc__= """
TALAN
=====
A Technical Algorithmic Analytics Python Package
"""
__version__ = "0.1.4"
__author__ = 'Ted Hong'
__credits__ = 'Beyondbond Risk Lab'
__name__ = 'talan'
from .talan_wrap import *
__all__= list(talan_vars.keys())
